import datetime
import itertools
import os
import socket
import time

from queue import Empty as QueueEmpty
from queue import Queue
from threading import Thread
from unittest import mock

import pytest
import webtest

from gevent.server import StreamServer

from baseplate.observers.timeout import ServerTimeout
from baseplate.server.queue_consumer import HealthcheckApp
from baseplate.server.queue_consumer import MessageHandler
from baseplate.server.queue_consumer import PumpWorker
from baseplate.server.queue_consumer import QueueConsumer
from baseplate.server.queue_consumer import QueueConsumerFactory
from baseplate.server.queue_consumer import QueueConsumerServer


pytestmark = pytest.mark.skipif(
    "CI" not in os.environ, reason="tests takes too long to run for normal local iteration"
)


class FakePumpWorker(PumpWorker):
    def __init__(self, work_queue, raises=None):
        self.work_queue = work_queue
        self.started = False
        self.stopped = False
        self.raises = raises

    def run(self):
        self.started = True
        if self.raises is not None:
            raise self.raises
        for i in itertools.count():
            self.work_queue.put(i)
            if self.stopped:
                break
            else:
                time.sleep(0.1)

    def stop(self):
        self.stopped = True


class FakeMessageHandler(MessageHandler):
    def __init__(self, raises=None):
        self.raises = raises
        self.messages = []

    def handle(self, message):
        if self.raises is not None:
            raise self.raises
        self.messages.append(message)
        if len(self.messages) > 10:
            self.messages = self.messages[1:10]


class FakeQueueConsumerFactory(QueueConsumerFactory):
    def __init__(self, pump_raises=None, handler_raises=None):
        self.pump_raises = pump_raises
        self.handler_raises = handler_raises

    def build_pump_worker(self, work_queue):
        return FakePumpWorker(work_queue, raises=self.pump_raises)

    def build_message_handler(self):
        return FakeMessageHandler(raises=self.handler_raises)

    def build_health_checker(self, listener):
        return mock.Mock(spec=StreamServer)


class TestQueueConsumer:
    @pytest.fixture
    def consumer(self):
        c = QueueConsumer(Queue(maxsize=5), FakeMessageHandler())
        c._queue_timeout = 0.1
        yield c
        try:
            c.stop()
        except AssertionError:
            pass

    def test_stop_before_run(self, consumer):
        with pytest.raises(AssertionError):
            consumer.stop()

    def test_run(self, consumer):
        for i in range(5):
            consumer.work_queue.put(i)
        t = Thread(target=consumer.run, daemon=True)
        t.start()
        time.sleep(0.1)
        assert consumer.message_handler.messages == [i for i in range(5)]
        assert t.is_alive()
        with pytest.raises(AssertionError):
            consumer.run()

    def test_stop(self, consumer):
        t = Thread(target=consumer.run)
        t.start()
        time.sleep(0.1)
        assert t.is_alive()
        consumer.stop()
        time.sleep(0.5)
        t.join()
        assert consumer.started
        assert consumer.stopped
        assert not t.is_alive()
        with pytest.raises(AssertionError):
            consumer.stop()

    def test_handler_error_stops(self, consumer):
        consumer.work_queue.put(0)
        consumer.message_handler.raises = Exception()
        with pytest.raises(Exception):
            consumer.run()

    def test_queue_empty_in_handler_error(self, consumer):
        # QueueConsumer.run sets a timeout when checking it's work queue so it
        # can periodically check if it has been stopped.  We want to be sure we
        # only catch `queue.Empty` in that case, but let `queue.Empty` errors
        # thrown by the MessageHandler bubble up.
        consumer.work_queue.put(0)
        consumer.message_handler.raises = QueueEmpty()
        with pytest.raises(QueueEmpty):
            consumer.run()


class TestQueueConsumerServer:
    @pytest.fixture
    def build_server(self):
        server = [None]

        def _build_server(max_concurrency=5, pump_raises=None, handler_raises=None):
            server[0] = QueueConsumerServer.new(
                consumer_factory=FakeQueueConsumerFactory(
                    pump_raises=pump_raises, handler_raises=handler_raises
                ),
                max_concurrency=max_concurrency,
                listener=mock.Mock(spec=socket.socket),
                stop_timeout=datetime.timedelta(seconds=30),
            )
            return server[0]

        yield _build_server

        try:
            if server[0] is not None:
                server[0].stop()
        except AssertionError:
            pass

    @pytest.fixture
    def server(self, build_server):
        return build_server()

    def test_new(self):
        max_concurrency = 5
        server = QueueConsumerServer.new(
            consumer_factory=FakeQueueConsumerFactory(),
            max_concurrency=max_concurrency,
            listener=mock.Mock(spec=socket.socket),
            stop_timeout=datetime.timedelta(seconds=30),
        )
        assert not server.started
        assert not server.stopped
        assert not server.pump.started
        assert not server.pump.stopped
        assert server.pump.work_queue.maxsize == 7
        server.healthcheck_server.start.assert_not_called()
        server.healthcheck_server.stop.assert_not_called()
        assert len(server.handlers) == max_concurrency
        assert len(server.handlers) == len(server.threads)
        for handler in server.handlers:
            assert not handler.started
            assert not handler.stopped

    def test_start(self, server):
        server.start()
        assert server.started
        assert not server.stopped
        assert server.pump.started
        assert not server.pump.stopped
        server.healthcheck_server.start.assert_called_once()
        server.healthcheck_server.stop.assert_not_called()
        for handler in server.handlers:
            assert handler.started
            assert not handler.stopped
        with pytest.raises(AssertionError):
            server.start()

    def test_stop(self, server):
        server.start()
        server.stop()
        assert server.started
        assert server.stopped
        assert server.pump.started
        assert server.pump.stopped
        server.healthcheck_server.start.assert_called_once()
        server.healthcheck_server.stop.assert_called_once()
        for handler in server.handlers:
            assert handler.started
            assert handler.stopped
        with pytest.raises(AssertionError):
            server.stop()

    def test_stop_before_start(self, server):
        with pytest.raises(AssertionError):
            server.stop()

    def test_pump_exception_terminates(self, build_server):
        server = build_server(max_concurrency=1, pump_raises=Exception())
        server._terminate = mock.Mock()
        server.start()
        time.sleep(0.5)
        server._terminate.assert_called_once()

    def test_handler_exception_terminates(self, build_server):
        server = build_server(max_concurrency=1, handler_raises=Exception())
        server._terminate = mock.Mock()
        server.start()
        time.sleep(0.5)
        server._terminate.assert_called_once()

    def test_handler_timeout_terminates(self, build_server):
        server = build_server(max_concurrency=1, handler_raises=ServerTimeout("", 10, False))
        server._terminate = mock.Mock()
        server.start()
        time.sleep(0.5)
        server._terminate.assert_called_once()


def test_healthcheck():
    healthcheck_app = HealthcheckApp()
    test_app = webtest.TestApp(healthcheck_app)

    response = test_app.get("/health")
    assert response.content_type == "application/json"
    assert isinstance(response.json, dict)



import contextlib
import unittest

try:
    import redis
except ImportError:
    raise unittest.SkipTest("redis-py is not installed")

from baseplate.clients.redis import ACTIVE_REQUESTS
from baseplate.clients.redis import REQUESTS_TOTAL
from baseplate.clients.redis import LATENCY_SECONDS
from baseplate.clients.redis import RedisClient
from baseplate import Baseplate

from . import TestBaseplateObserver, get_endpoint_or_skip_container

from baseplate.clients.redis import MessageQueue
from baseplate.lib.message_queue import TimedOutError
from prometheus_client import REGISTRY

redis_endpoint = get_endpoint_or_skip_container("redis", 6379)


class RedisIntegrationTests(unittest.TestCase):
    def setUp(self):
        self.setup_baseplate_redis()

    def setup_baseplate_redis(self, redis_client_kwargs={}):
        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate({"redis.url": f"redis://{redis_endpoint}/0"})
        baseplate.register(self.baseplate_observer)
        baseplate.configure_context({"redis": RedisClient(**redis_client_kwargs)})

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple_command(self):
        with self.server_span:
            result = self.context.redis.ping()

        self.assertTrue(result)

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "redis.PING")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)

    def test_error(self):
        with self.server_span:
            with self.assertRaises(redis.ResponseError):
                self.context.redis.execute_command("crazycommand")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNotNone(span_observer.on_finish_exc_info)

    def test_lock(self):
        with self.server_span:
            with self.context.redis.lock("foo"):
                pass

        server_span_observer = self.baseplate_observer.get_only_child()

        self.assertGreater(len(server_span_observer.children), 0)
        for span_observer in server_span_observer.children:
            self.assertTrue(span_observer.on_start_called)
            self.assertTrue(span_observer.on_finish_called)

    def test_pipeline(self):
        with self.server_span:
            with self.context.redis.pipeline("foo") as pipeline:
                pipeline.ping()
                pipeline.execute()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "redis.pipeline_foo")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)

    def test_metrics(self):
        client_name = "redisclient"
        for client_name_kwarg_name in [
            "redis_client_name",
            "client_name",
        ]:
            with self.subTest():
                self.setup_baseplate_redis(
                    redis_client_kwargs={
                        client_name_kwarg_name: client_name,
                    },
                )
                # Clear preometheus metrics
                ACTIVE_REQUESTS.clear()
                REQUESTS_TOTAL.clear()
                LATENCY_SECONDS.clear()
                expected_labels = {
                    "redis_client_name": client_name,
                    "redis_type": "standalone",
                    "redis_command": "SET",
                    "redis_database": "0",
                }
                with self.server_span:
                    self.context.redis.set("prometheus", "rocks")

                request_labels = {**expected_labels, "redis_success": "true"}
                assert (
                    REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", request_labels)
                    == 1.0
                ), "Unexpected value for REQUESTS_TOTAL metric. Expected one 'set' command"
                assert (
                    REGISTRY.get_sample_value(
                        f"{LATENCY_SECONDS._name}_bucket", {**request_labels, "le": "+Inf"}
                    )
                    == 1.0
                ), "Expected one 'set' latency request"
                assert (
                    REGISTRY.get_sample_value(
                        ACTIVE_REQUESTS._name, {**expected_labels, "redis_type": "standalone"}
                    )
                    == 0.0
                ), "Should have 0 (and not None) active requests"


class RedisMessageQueueTests(unittest.TestCase):
    qname = "redisTestQueue"

    def setUp(self):
        self.pool = redis.ConnectionPool(
            host=redis_endpoint.address.host, port=redis_endpoint.address.port
        )

    def test_put_get(self):
        message_queue = MessageQueue(self.qname, self.pool)

        with contextlib.closing(message_queue) as mq:
            mq.put(b"x")
            message = mq.get()
            self.assertEqual(message, b"x")

    def test_get_zero_timeout(self):
        message_queue = MessageQueue(self.qname, self.pool)

        message_queue.put(b"y")
        message = message_queue.get(timeout=0)
        self.assertEqual(message, b"y")

        with self.assertRaises(TimedOutError):
            message_queue.get(timeout=0)

    def test_put_zero_timeout(self):
        message_queue = MessageQueue(self.qname, self.pool)

        with contextlib.closing(message_queue) as mq:
            mq.put(b"x", timeout=0)
            message = mq.get()
            self.assertEqual(message, b"x")

    def tearDown(self):
        redis.Redis(connection_pool=self.pool).delete(self.qname)



import base64
import unittest

from unittest import mock

from pyramid.response import Response

from baseplate import Baseplate
from baseplate import BaseplateObserver
from baseplate import ServerSpanObserver

from . import FakeEdgeContextFactory


try:
    import webtest

    from baseplate.frameworks.pyramid import BaseplateConfigurator
    from baseplate.frameworks.pyramid import ServerSpanInitialized
    from baseplate.frameworks.pyramid import StaticTrustHandler
    from pyramid.config import Configurator
    from pyramid.httpexceptions import HTTPInternalServerError
except ImportError:
    raise unittest.SkipTest("pyramid/webtest is not installed")


class TestException(Exception):
    pass


class ControlFlowException(Exception):
    pass


class ControlFlowException2(Exception):
    pass


class ExceptionViewException(Exception):
    pass


def example_application(request):
    if "error" in request.params:
        raise TestException("this is a test")

    if "control_flow_exception" in request.params:
        raise ControlFlowException()

    if "exception_view_exception" in request.params:
        raise ControlFlowException2()

    if "stream" in request.params:

        def make_iter():
            yield b"foo"
            yield b"bar"

        return Response(status_code=200, app_iter=make_iter())

    return {"test": "success"}


def render_exception_view(request):
    return HTTPInternalServerError(title="a fancy title", body="a fancy explanation")


def render_bad_exception_view(request):
    raise ExceptionViewException()


def local_tracing_within_context(request):
    with request.span.make_child("local-req", local=True, component_name="in-context"):
        pass
    return {"trace": "success"}


class ConfiguratorTests(unittest.TestCase):
    def setUp(self):
        configurator = Configurator()
        configurator.add_route("example", "/example", request_method="GET")
        configurator.add_route("route", "/route/{hello}/world", request_method="GET")
        configurator.add_route("trace_context", "/trace_context", request_method="GET")

        configurator.add_view(example_application, route_name="example", renderer="json")
        configurator.add_view(example_application, route_name="route", renderer="json")

        configurator.add_view(
            local_tracing_within_context, route_name="trace_context", renderer="json"
        )

        configurator.add_view(render_exception_view, context=ControlFlowException, renderer="json")

        configurator.add_view(
            render_bad_exception_view, context=ControlFlowException2, renderer="json"
        )

        self.observer = mock.Mock(spec=BaseplateObserver)
        self.server_observer = mock.Mock(spec=ServerSpanObserver)

        def _register_mock(context, server_span):
            server_span.register(self.server_observer)

        self.observer.on_server_span_created.side_effect = _register_mock

        self.baseplate = Baseplate()
        self.baseplate.register(self.observer)
        self.baseplate_configurator = BaseplateConfigurator(
            self.baseplate,
            edge_context_factory=FakeEdgeContextFactory(),
            header_trust_handler=StaticTrustHandler(trust_headers=True),
        )
        configurator.include(self.baseplate_configurator.includeme)
        self.context_init_event_subscriber = mock.Mock()
        configurator.add_subscriber(self.context_init_event_subscriber, ServerSpanInitialized)
        app = configurator.make_wsgi_app()
        self.test_app = webtest.TestApp(app)

    @mock.patch("random.getrandbits")
    def test_no_trace_headers(self, getrandbits):
        getrandbits.return_value = 1234
        self.test_app.get("/example")

        self.assertEqual(self.observer.on_server_span_created.call_count, 1)

        context, server_span = self.observer.on_server_span_created.call_args[0]
        self.assertEqual(server_span.trace_id, "1234")
        self.assertEqual(server_span.parent_id, None)
        self.assertEqual(server_span.id, "1234")

        self.assertTrue(self.server_observer.on_start.called)
        self.assertTrue(self.server_observer.on_finish.called)
        self.assertTrue(self.context_init_event_subscriber.called)

    def test_trace_headers(self):
        self.test_app.get(
            "/example",
            headers={
                "X-Trace": "1234",
                "X-Edge-Request": base64.b64encode(FakeEdgeContextFactory.RAW_BYTES).decode(),
                "X-Parent": "2345",
                "X-Span": "3456",
                "X-Sampled": "1",
                "X-Flags": "1",
            },
        )

        self.assertEqual(self.observer.on_server_span_created.call_count, 1)

        context, server_span = self.observer.on_server_span_created.call_args[0]
        self.assertEqual(server_span.trace_id, "1234")
        self.assertEqual(server_span.parent_id, "2345")
        self.assertEqual(server_span.id, "3456")
        self.assertEqual(server_span.sampled, True)
        self.assertEqual(server_span.flags, 1)

        self.assertTrue(self.server_observer.on_start.called)
        self.assertTrue(self.server_observer.on_finish.called)
        self.assertTrue(self.context_init_event_subscriber.called)

    def test_edge_request_headers(self):
        self.test_app.get(
            "/example",
            headers={
                "X-Trace": "1234",
                "X-Edge-Request": base64.b64encode(FakeEdgeContextFactory.RAW_BYTES).decode(),
                "X-Parent": "2345",
                "X-Span": "3456",
                "X-Sampled": "1",
                "X-Flags": "1",
            },
        )
        context, _ = self.observer.on_server_span_created.call_args[0]
        assert context.edge_context == FakeEdgeContextFactory.DECODED_CONTEXT

    def test_empty_edge_request_headers(self):
        self.test_app.get(
            "/example",
            headers={
                "X-Trace": "1234",
                "X-Edge-Request": "",
                "X-Parent": "2345",
                "X-Span": "3456",
                "X-Sampled": "1",
                "X-Flags": "1",
            },
        )
        context, _ = self.observer.on_server_span_created.call_args[0]
        self.assertEqual(context.raw_edge_context, b"")

    def test_not_found(self):
        self.test_app.get("/nope", status=404)

        self.assertFalse(self.observer.on_server_span_created.called)
        self.assertFalse(self.context_init_event_subscriber.called)

    def test_exception_caught(self):
        with self.assertRaises(TestException):
            self.test_app.get("/example?error")

        self.assertTrue(self.server_observer.on_start.called)
        self.assertTrue(self.server_observer.on_finish.called)
        self.assertTrue(self.context_init_event_subscriber.called)
        _, captured_exc, _ = self.server_observer.on_finish.call_args[0][0]
        self.assertIsInstance(captured_exc, TestException)

    def test_control_flow_exception_not_caught(self):
        response = self.test_app.get("/example?control_flow_exception", status=500)

        self.assertTrue(self.server_observer.on_start.called)
        self.assertTrue(self.server_observer.on_finish.called)
        self.assertTrue(self.context_init_event_subscriber.called)
        self.assertTrue(b"a fancy explanation", response.body)
        args, _ = self.server_observer.on_finish.call_args
        self.assertEqual(args[0], None)

    def test_exception_in_exception_view_caught(self):
        with self.assertRaises(ExceptionViewException):
            self.test_app.get("/example?exception_view_exception")

        self.assertTrue(self.server_observer.on_start.called)
        self.assertTrue(self.server_observer.on_finish.called)
        self.assertTrue(self.context_init_event_subscriber.called)
        _, captured_exc, _ = self.server_observer.on_finish.call_args[0][0]
        self.assertIsInstance(captured_exc, ExceptionViewException)

    @mock.patch("random.getrandbits")
    def test_distrust_headers(self, getrandbits):
        getrandbits.return_value = 9999
        self.baseplate_configurator.header_trust_handler.trust_headers = False

        self.test_app.get(
            "/example", headers={"X-Trace": "1234", "X-Parent": "2345", "X-Span": "3456"}
        )

        context, server_span = self.observer.on_server_span_created.call_args[0]
        self.assertEqual(server_span.trace_id, str(getrandbits.return_value))
        self.assertEqual(server_span.parent_id, None)
        self.assertEqual(server_span.id, str(getrandbits.return_value))

    def test_local_trace_in_context(self):
        self.test_app.get("/trace_context")
        self.assertEqual(self.server_observer.on_child_span_created.call_count, 1)
        child_span = self.server_observer.on_child_span_created.call_args[0][0]
        context, server_span = self.observer.on_server_span_created.call_args[0]
        self.assertNotEqual(child_span.context, context)

    def test_streaming_response(self):
        class StreamingTestResponse(webtest.TestResponse):
            def decode_content(self):
                # keep your grubby hands off the app_iter, webtest!!!!
                pass

            @property
            def body(self):
                # seriously
                pass

        class StreamingTestRequest(webtest.TestRequest):
            ResponseClass = StreamingTestResponse

        self.test_app.RequestClass = StreamingTestRequest

        response = self.test_app.get("/example?stream")

        # ok, we've returned from the wsgi app but the iterator's not done
        # so... we should have started the span but not finished it yet
        self.assertTrue(self.server_observer.on_start.called)
        self.assertFalse(self.server_observer.on_finish.called)

        self.assertEqual(b"foo", next(response.app_iter))
        self.assertFalse(self.server_observer.on_finish.called)

        self.assertEqual(b"bar", next(response.app_iter))
        self.assertFalse(self.server_observer.on_finish.called)

        with self.assertRaises(StopIteration):
            next(response.app_iter)
        self.assertTrue(self.server_observer.on_finish.called)

        response.app_iter.close()



"""Watch a file and keep a parsed copy in memory that's updated on changes.

The contents of the file are re-loaded and parsed only when necessary.

For example, a JSON file like the following:

.. highlight:: ini

.. include:: ../../../watched_file.json
   :literal:

.. testsetup::

    import json
    from baseplate.lib.file_watcher import FileWatcher
    path = "docs/watched_file.json"

might be watched and parsed like this:

.. highlight:: py

.. doctest::

    >>> watcher = FileWatcher(path, parser=json.load)
    >>> watcher.get_data() == {u"one": 1, u"two": 2}
    True

The return value of :py:meth:`~baseplate.lib.file_watcher.FileWatcher.get_data`
would change whenever the underlying file changes.

"""
import logging
import os
import typing

from typing import Callable
from typing import Generic
from typing import IO
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from baseplate.lib.retry import RetryPolicy


logger = logging.getLogger(__name__)

DEFAULT_FILEWATCHER_BACKOFF = 0.01


class _NOT_LOADED:
    pass


class WatchedFileNotAvailableError(Exception):
    """Raised when the watched file could not be loaded."""

    def __init__(self, path: str, inner: Union[Exception, str]):
        super().__init__(f"{path}: {inner}")
        self.path = path
        self.inner = inner


T = TypeVar("T")


class _OpenOptions(NamedTuple):
    mode: str
    encoding: Optional[str]
    newline: Optional[str]


class FileWatcher(Generic[T]):
    r"""Watch a file and load its data when it changes.

    :param path: Full path to a file to watch.
    :param parser: A callable that takes an open file object, parses
        or otherwise interprets the file, and returns whatever data is meaningful.
    :param timeout: How long, in seconds, to block instantiation
        waiting for the watched file to become available (defaults to not blocking).
    :param binary: Should the file be opened in binary mode. If
        `True` the file will be opened with the mode `"rb"`, otherwise it will be
        opened with the mode `"r"`. (defaults to `"r"`)
    :param encoding: The name of the encoding used to decode the file. The
        default encoding is platform dependent (whatever
        :py:func:`locale.getpreferredencoding` returns), but any text encoding
        supported by Python can be used.  This is not supported if `binary` is
        set to `True`.
    :param newline: Controls how universal newlines mode works
        (it only applies to text mode). It can be `None`, `""`, `"\\n"`, `"\\r"`,
        and `"\\r\\n"`.  This is not supported if `binary` is set to `True`.
    :param backoff: retry backoff time for the file watcher. Defaults to
        None, which is mapped to DEFAULT_FILEWATCHER_BACKOFF.

    """

    def __init__(
        self,
        path: str,
        parser: Callable[[IO], T],
        timeout: Optional[float] = None,
        binary: bool = False,
        encoding: Optional[str] = None,
        newline: Optional[str] = None,
        backoff: Optional[float] = None,
    ):
        if binary and encoding is not None:
            raise TypeError("'encoding' is not supported in binary mode.")

        if binary and newline is not None:
            raise TypeError("'newline' is not supported in binary mode.")

        self._path = path
        self._parser = parser
        self._mtime = 0.0
        self._data: Union[T, Type[_NOT_LOADED]] = _NOT_LOADED
        self._open_options = _OpenOptions(
            mode="rb" if binary else "r",
            encoding=encoding or ("UTF-8" if not binary else None),
            newline=newline,
        )

        backoff = backoff or DEFAULT_FILEWATCHER_BACKOFF

        if timeout is not None:
            last_error = None
            for _ in RetryPolicy.new(budget=timeout, backoff=backoff):
                if self._data is not _NOT_LOADED:
                    break

                try:
                    self.get_data()
                except WatchedFileNotAvailableError as exc:
                    last_error = exc
                else:
                    break

                logging.warning("%s: file not yet available. sleeping.", path)
            else:
                last_error = typing.cast(WatchedFileNotAvailableError, last_error)
                raise WatchedFileNotAvailableError(
                    self._path, f"timed out. last error was: {last_error.inner}"
                )

    def get_data(self) -> T:
        """Return the current contents of the file, parsed.

        The watcher ensures that the file is re-loaded and parsed whenever its
        contents change. Parsing only occurs when necessary, not on each call
        to this method. This method returns whatever the most recent call to
        the parser returned.

        Make sure to call this method each time you need data from the file
        rather than saving its results elsewhere. This ensures you always have
        the freshest data.

        """
        return self.get_data_and_mtime()[0]

    def get_data_and_mtime(self) -> Tuple[T, float]:
        """Return tuple of the current contents of the file and file mtime.

        The watcher ensures that the file is re-loaded and parsed whenever its
        contents change. Parsing only occurs when necessary, not on each call
        to this method. This method returns whatever the most recent call to
        the parser returned.

        When file content was changed, it returns the recent mtime,
        notify the caller the content is different from previous cached.

        Make sure to call this method each time you need data from the file
        rather than saving its results elsewhere. This ensures you always have
        the freshest data.

        """
        try:
            current_mtime = os.path.getmtime(self._path)
        except OSError as exc:
            if self._data is _NOT_LOADED:
                raise WatchedFileNotAvailableError(self._path, exc)
            return typing.cast(T, self._data), self._mtime

        if self._mtime < current_mtime:
            logger.debug("Loading %s.", self._path)
            try:
                # pylint: disable=unspecified-encoding
                with open(self._path, **self._open_options._asdict()) as f:
                    self._data = self._parser(f)
            except Exception as exc:
                if self._data is _NOT_LOADED:
                    raise WatchedFileNotAvailableError(self._path, exc)
                logger.warning("%s: failed to load, using cached data: %s", self._path, exc)
            self._mtime = current_mtime

        return typing.cast(T, self._data), self._mtime



import unittest

from baseplate.lib.secrets import CorruptSecretError
from baseplate.lib.secrets import CredentialSecret
from baseplate.lib.secrets import DirectorySecretsStore
from baseplate.lib.secrets import parse_vault_csi
from baseplate.lib.secrets import SecretNotFoundError
from baseplate.lib.secrets import secrets_store_from_config
from baseplate.testing.lib.file_watcher import FakeFileWatcher


class StoreDirectoryTests(unittest.TestCase):
    def setUp(self):
        self.fake_filewatcher_1 = FakeFileWatcher()
        self.fake_filewatcher_2 = FakeFileWatcher()
        self.fake_filewatcher_3 = FakeFileWatcher()
        self.store = DirectorySecretsStore("/whatever", parse_vault_csi)
        self.store._filewatchers["secret1"] = self.fake_filewatcher_1
        self.store._filewatchers["secret2"] = self.fake_filewatcher_2
        self.store._filewatchers["secret3"] = self.fake_filewatcher_3

    def test_file_not_found(self):
        with self.assertRaises(SecretNotFoundError):
            self.store.get_raw("test")

    def test_vault_info(self):
        with self.assertRaises(NotImplementedError):
            self.store.get_vault_token()

        with self.assertRaises(NotImplementedError):
            self.store.get_vault_url()

    def test_raw_secrets(self):
        self.fake_filewatcher_1.data = {
            "data": {"something": "exists"},
        }

        self.assertEqual(self.store.get_raw("secret1"), {"something": "exists"})

        with self.assertRaises(SecretNotFoundError):
            self.store.get_raw("secret0")

    def test_simple_secrets(self):
        # simple test
        self.fake_filewatcher_1.data = {
            "data": {"type": "simple", "value": "easy"},
        }
        self.assertEqual(self.store.get_simple("secret1"), b"easy")

        # test base64
        self.fake_filewatcher_2.data = {
            "data": {"type": "simple", "value": "aHVudGVyMg==", "encoding": "base64"},
        }
        self.assertEqual(self.store.get_simple("secret2"), b"hunter2")

        # test unknown encoding
        self.fake_filewatcher_3.data = {
            "data": {"type": "simple", "value": "sdlfkj", "encoding": "mystery"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("secret3")

        # test not simple
        self.fake_filewatcher_1.data = {
            "data": {"something": "else"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("secret1")

        # test no value
        self.fake_filewatcher_2.data = {
            "data": {"type": "simple"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("secret2")

        # test bad base64
        self.fake_filewatcher_3.data = {
            "data": {"type": "simple", "value": "aHVudGVyMg", "encoding": "base64"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("secret3")

    def test_versioned_secrets(self):
        # simple test
        self.fake_filewatcher_1.data = {
            "data": {"type": "versioned", "current": "easy"},
        }
        simple = self.store.get_versioned("secret1")
        self.assertEqual(simple.current, b"easy")
        self.assertEqual(list(simple.all_versions), [b"easy"])

        # test base64
        self.fake_filewatcher_2.data = {
            "data": {
                "type": "versioned",
                "previous": "aHVudGVyMQ==",
                "current": "aHVudGVyMg==",
                "next": "aHVudGVyMw==",
                "encoding": "base64",
            },
        }
        encoded = self.store.get_versioned("secret2")
        self.assertEqual(encoded.previous, b"hunter1")
        self.assertEqual(encoded.current, b"hunter2")
        self.assertEqual(encoded.next, b"hunter3")
        self.assertEqual(list(encoded.all_versions), [b"hunter2", b"hunter1", b"hunter3"])

        # test unknown encoding
        self.fake_filewatcher_3.data = {
            "data": {"type": "versioned", "current": "sdlfkj", "encoding": "mystery"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("secret3")

        # test not versioned
        self.fake_filewatcher_1.data = {
            "data": {"something": "else"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("secret1")

        # test no value
        self.fake_filewatcher_2.data = {
            "data": {"type": "versioned"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("secret2")

        # test bad base64
        self.fake_filewatcher_3.data = {
            "data": {"type": "simple", "value": "aHVudGVyMg", "encoding": "base64"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("secret3")

    def test_credential_secrets(self):
        # simple test
        self.fake_filewatcher_1.data = {
            "data": {"type": "credential", "username": "user", "password": "password"},
        }
        self.assertEqual(
            self.store.get_credentials("secret1"), CredentialSecret("user", "password")
        )

        # test identiy
        self.fake_filewatcher_2.data = {
            "data": {
                "type": "credential",
                "username": "spez",
                "password": "hunter2",
                "encoding": "identity",
            },
        }
        self.assertEqual(self.store.get_credentials("secret2"), CredentialSecret("spez", "hunter2"))

        # test base64
        self.fake_filewatcher_2.data = {
            "data": {
                "type": "credential",
                "username": "foo",
                "password": "aHVudGVyMg==",
                "encoding": "base64",
            },
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret2")

        # test unknkown encoding
        self.fake_filewatcher_3.data = {
            "data": {
                "type": "credential",
                "username": "fizz",
                "password": "buzz",
                "encoding": "something",
            },
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret3")

        # test not credentials
        self.fake_filewatcher_1.data = {
            "data": {"type": "versioned", "current": "easy"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret1")

        # test no values
        self.fake_filewatcher_2.data = {
            "data": {"type": "credential"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret2")

        # test no username
        self.fake_filewatcher_3.data = {
            "data": {"type": "credential", "password": "password"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret3")

        # test no password
        self.fake_filewatcher_1.data = {
            "data": {"type": "credential", "username": "user"},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret1")

        # test bad type
        self.fake_filewatcher_2.data = {
            "data": {"type": "credential", "username": "user", "password": 100},
        }
        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("secret2")


class StoreFromConfigTests(unittest.TestCase):
    def test_make_store(self):
        secrets = secrets_store_from_config(
            {"secrets.path": "/tmp", "secrets.provider": "vault_csi"}
        )
        self.assertIsInstance(secrets, DirectorySecretsStore)



"""Components for processing Baseplate spans for service request tracing."""
import collections
import json
import logging
import queue
import random
import socket
import threading
import time
import typing

from datetime import datetime
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional

import requests

from requests.exceptions import RequestException

from baseplate import _ExcInfo
from baseplate import BaseplateObserver
from baseplate import LocalSpan
from baseplate import RequestContext
from baseplate import Span
from baseplate import SpanObserver
from baseplate.lib import config
from baseplate.lib import warn_deprecated
from baseplate.lib.message_queue import MessageQueue
from baseplate.lib.message_queue import TimedOutError
from baseplate.observers.timeout import ServerTimeout


if typing.TYPE_CHECKING:
    SpanQueue = queue.Queue["TraceSpanObserver"]  # pylint: disable=unsubscriptable-object
else:
    SpanQueue = queue.Queue


logger = logging.getLogger(__name__)

# Suppress noisy INFO logging of underlying connection management module
logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)

# Span annotation types
ANNOTATIONS = {
    "CLIENT_SEND": "cs",
    "CLIENT_RECEIVE": "cr",
    "SERVER_SEND": "ss",
    "SERVER_RECEIVE": "sr",
    "LOCAL_COMPONENT": "lc",
    "COMPONENT": "component",
    "DEBUG": "debug",
    "ERROR": "error",
}

# Feature flags
FLAGS = {
    # Ensures the trace passes ALL samplers
    "DEBUG": 1
}


# Max size for a string representation of a span when recorded to a POSIX queue
MAX_SPAN_SIZE = 102400
# Max number of spans allowed in POSIX queue at one time
MAX_QUEUE_SIZE = 10000


def current_epoch_microseconds() -> int:
    """Return current UTC time since epoch in microseconds."""
    epoch_ts = datetime.utcfromtimestamp(0)
    return int((datetime.utcnow() - epoch_ts).total_seconds() * 1000 * 1000)


class TracingClient(NamedTuple):
    service_name: str
    sample_rate: float
    recorder: "Recorder"


def make_client(
    service_name: str,
    tracing_endpoint: Optional[config.EndpointConfiguration] = None,
    tracing_queue_name: Optional[str] = None,
    max_span_queue_size: int = 50000,
    num_span_workers: int = 5,
    span_batch_interval: float = 0.5,
    num_conns: int = 100,
    sample_rate: float = 0.1,
    log_if_unconfigured: bool = True,
) -> TracingClient:
    """Create and return a tracing client based on configuration options.

    This client can be used by the :py:class:`TraceBaseplateObserver`.

    :param service_name: The name for the service this observer
        is registered to.
    :param tracing_endpoint: Destination to record span data.
    :param tracing_queue_name: POSIX queue name for reporting spans.
    :param num_conns: pool size for remote recorder connection pool.
    :param max_span_queue_size: span processing queue limit.
    :param num_span_workers: number of worker threads for span processing.
    :param span_batch_interval: wait time for span processing in seconds.
    :param sample_rate: percentage of unsampled requests to record traces for.
    """
    recorder: Recorder
    if tracing_queue_name:
        recorder = SidecarRecorder(tracing_queue_name)
    elif tracing_endpoint:
        warn_deprecated("In-app trace publishing is deprecated in favor of the sidecar model.")
        remote_addr = str(tracing_endpoint.address)
        recorder = RemoteRecorder(
            remote_addr,
            num_conns=num_conns,
            max_queue_size=max_span_queue_size,
            num_workers=num_span_workers,
            batch_wait_interval=span_batch_interval,
        )
    elif log_if_unconfigured:
        recorder = LoggingRecorder(
            max_queue_size=max_span_queue_size,
            num_workers=num_span_workers,
            batch_wait_interval=span_batch_interval,
        )
    else:
        recorder = NullRecorder(
            max_queue_size=max_span_queue_size,
            num_workers=num_span_workers,
            batch_wait_interval=span_batch_interval,
        )

    return TracingClient(service_name, sample_rate, recorder)


class TraceBaseplateObserver(BaseplateObserver):
    """Distributed tracing observer.

    This observer handles Zipkin-compatible distributed tracing
    instrumentation for both inbound and outbound requests.
    Baseplate span-specific tracing observers (TraceSpanObserver
    and TraceServerSpanObserver) are registered for tracking,
    serializing, and recording span data.

    :param baseplate.observers.tracing.TracingClient client: The client
        where metrics will be sent.

    """

    def __init__(self, tracing_client: TracingClient):
        self.service_name = tracing_client.service_name
        self.sample_rate = tracing_client.sample_rate
        self.recorder = tracing_client.recorder
        try:
            self.hostname = socket.gethostbyname(socket.gethostname())
        except socket.gaierror as e:
            logger.warning("Hostname could not be resolved, error=%s", e)
            self.hostname = "undefined"

    @classmethod
    def force_sampling(cls, span: Span) -> bool:
        if span.flags:
            return span.flags & FLAGS["DEBUG"] == FLAGS["DEBUG"]
        return False

    def should_sample(self, span: Span) -> bool:
        should_sample = False
        if span.sampled is None:
            should_sample = random.random() < self.sample_rate
        else:
            should_sample = span.sampled
        return should_sample or self.force_sampling(span)

    def on_server_span_created(self, context: RequestContext, server_span: Span) -> None:
        if self.should_sample(server_span):
            server_span.sampled = True
            observer = TraceServerSpanObserver(
                self.service_name, self.hostname, server_span, self.recorder
            )
            server_span.register(observer)
        else:
            server_span.sampled = False


class TraceSpanObserver(SpanObserver):
    """Span recording observer for outgoing request child spans.

    This observer implements the client-side span portion of a
    Zipkin request trace.
    """

    def __init__(self, service_name: str, hostname: str, span: Span, recorder: "Recorder"):
        self.service_name = service_name
        self.hostname = hostname
        self.recorder = recorder
        self.span = span
        self.start: Optional[int] = None
        self.end: Optional[int] = None
        self.elapsed: Optional[int] = None
        self.binary_annotations: List[Dict[str, Any]] = []
        self.counters: DefaultDict[str, float] = collections.defaultdict(float)
        self.on_set_tag(ANNOTATIONS["COMPONENT"], "baseplate")
        super().__init__()

    def on_start(self) -> None:
        self.start = current_epoch_microseconds()
        self.client_send = self.start

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        if exc_info:
            self.on_set_tag(ANNOTATIONS["ERROR"], True)

        # Set a debug annotation for downstream processing to
        # utilize in span filtering.
        if self.span.flags and (self.span.flags & FLAGS["DEBUG"]):
            self.on_set_tag(ANNOTATIONS["DEBUG"], True)

        self.end = current_epoch_microseconds()
        self.elapsed = self.end - typing.cast(int, self.start)

        for key, value in self.counters.items():
            self.binary_annotations.append(self._create_binary_annotation(f"counter.{key}", value))

        self.recorder.send(self)

    def on_set_tag(self, key: str, value: Any) -> None:
        """Translate set tags to tracing binary annotations.

        Number-type values are coerced to strings.
        """
        self.binary_annotations.append(self._create_binary_annotation(key, value))

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.counters[key] += delta

    def _endpoint_info(self) -> Dict[str, str]:
        return {"serviceName": self.service_name, "ipv4": self.hostname}

    def _create_time_annotation(self, annotation_type: str, timestamp: int) -> Dict[str, Any]:
        """Create Zipkin-compatible Annotation for a span.

        This should be used for generating span annotations with a time component,
        e.g. the core "cs", "cr", "ss", and "sr" Zipkin Annotations
        """
        return {"endpoint": self._endpoint_info(), "timestamp": timestamp, "value": annotation_type}

    def _create_binary_annotation(
        self, annotation_type: str, annotation_value: Any
    ) -> Dict[str, Any]:
        """Create Zipkin-compatible BinaryAnnotation for a span.

        This should be used for generating span annotations that
        do not have a time component, e.g. URI, arbitrary request tags
        """
        endpoint_info = self._endpoint_info()

        # Annotation values must be str type.
        if isinstance(annotation_value, bool):
            # only "lower" bool values so we aren't affecting any other types
            annotation_value = str(annotation_value).lower()
        elif not isinstance(annotation_value, str):
            annotation_value = str(annotation_value)

        return {"key": annotation_type, "value": annotation_value, "endpoint": endpoint_info}

    def _to_span_obj(
        self, annotations: List[Dict[str, Any]], binary_annotations: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        span = {
            "traceId": self.span.trace_id,
            "name": self.span.name,
            "id": self.span.id,
            "timestamp": self.start,
            "duration": self.elapsed,
            "annotations": annotations,
            "binaryAnnotations": binary_annotations,
        }

        span["parentId"] = self.span.parent_id or 0
        return span

    def _serialize(self) -> Dict[str, Any]:
        """Serialize span information into Zipkin-accepted format."""
        annotations = []

        annotations.append(
            self._create_time_annotation(ANNOTATIONS["CLIENT_SEND"], typing.cast(int, self.start))
        )

        annotations.append(
            self._create_time_annotation(ANNOTATIONS["CLIENT_RECEIVE"], typing.cast(int, self.end))
        )

        return self._to_span_obj(annotations, self.binary_annotations)


class TraceLocalSpanObserver(TraceSpanObserver):
    """Span recording observer for local spans.

    :param service_name: The name for the service this observer is registered
        to.
    :param component_name: The name for the local component the span for this
        observer is recording in.
    :param hostname: Name identifying the host of the service.
    :param span: Local span for this observer.
    :param recorder: Recorder for span trace.
    """

    def __init__(
        self,
        service_name: str,
        component_name: str,
        hostname: str,
        span: Span,
        recorder: "Recorder",
    ):
        self.component_name = component_name
        super().__init__(service_name, hostname, span, recorder)
        self.binary_annotations.append(
            self._create_binary_annotation(ANNOTATIONS["LOCAL_COMPONENT"], self.component_name)
        )

    def on_start(self) -> None:
        self.start = current_epoch_microseconds()

    def on_child_span_created(self, span: Span) -> None:
        """Perform tracing-related actions for child spans creation.

        Register new TraceSpanObserver for the child span
        being created so span start and finish get properly recorded.
        """
        trace_observer: TraceSpanObserver
        if isinstance(span, LocalSpan):
            trace_observer = TraceLocalSpanObserver(
                self.service_name,
                typing.cast(str, span.component_name),
                self.hostname,
                span,
                self.recorder,
            )

        else:
            trace_observer = TraceSpanObserver(
                self.service_name, self.hostname, span, self.recorder
            )
        span.register(trace_observer)

    def _serialize(self) -> Dict[str, Any]:
        return self._to_span_obj([], self.binary_annotations)


class TraceServerSpanObserver(TraceSpanObserver):
    """Span recording observer for incoming request spans.

    This observer implements the server-side span portion of a
    Zipkin request trace
    """

    def __init__(self, service_name: str, hostname: str, span: Span, recorder: "Recorder"):
        self.service_name = service_name
        self.span = span
        self.recorder = recorder
        super().__init__(service_name, hostname, span, recorder)

    def on_start(self) -> None:
        self.start = current_epoch_microseconds()

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        if exc_info and exc_info[0] is not None and issubclass(ServerTimeout, exc_info[0]):
            self.on_set_tag("timed_out", True)

        super().on_finish(exc_info)

    def on_child_span_created(self, span: Span) -> None:
        """Perform tracing-related actions for child spans creation.

        Register new TraceSpanObserver for the child span
        being created so span start and finish get properly recorded.
        """
        trace_observer: TraceSpanObserver
        if isinstance(span, LocalSpan):
            trace_observer = TraceLocalSpanObserver(
                self.service_name,
                typing.cast(str, span.component_name),
                self.hostname,
                span,
                self.recorder,
            )

        else:
            trace_observer = TraceSpanObserver(
                self.service_name, self.hostname, span, self.recorder
            )
        span.register(trace_observer)

    def _serialize(self) -> Dict[str, Any]:
        """Serialize span information into Zipkin-accepted format."""
        annotations: List[Dict[str, Any]] = []

        annotations.append(
            self._create_time_annotation(
                ANNOTATIONS["SERVER_RECEIVE"], typing.cast(int, self.start)
            )
        )
        annotations.append(
            self._create_time_annotation(ANNOTATIONS["SERVER_SEND"], typing.cast(int, self.end))
        )

        return self._to_span_obj(annotations, self.binary_annotations)


class Recorder:
    def send(self, span: TraceSpanObserver) -> None:
        raise NotImplementedError


class BaseBatchRecorder(Recorder):
    def __init__(
        self, max_queue_size: int, num_workers: int, max_span_batch: int, batch_wait_interval: float
    ):
        self.span_queue: SpanQueue = queue.Queue(maxsize=max_queue_size)
        self.batch_wait_interval = batch_wait_interval
        self.max_span_batch = max_span_batch
        self.logger = logging.getLogger(self.__class__.__name__)
        for _ in range(num_workers):
            self.flush_worker = threading.Thread(target=self._flush_spans)
            self.flush_worker.name = "span recorder"
            self.flush_worker.daemon = True
            self.flush_worker.start()

    def flush_func(self, spans: List[Dict[str, Any]]) -> None:
        raise NotImplementedError

    def _flush_spans(self) -> None:
        # This reads a batch of at most max_span_batch spans off the recorder queue
        # and sends them to a remote recording endpoint. If the queue
        # empties while being processed before reaching 10 spans, we flush
        # immediately.
        while True:
            spans: List[Dict[str, Any]] = []
            try:
                while len(spans) < self.max_span_batch:
                    spans.append(self.span_queue.get_nowait()._serialize())
            except queue.Empty:
                pass
            finally:
                if spans:
                    self.flush_func(spans)
                else:
                    time.sleep(self.batch_wait_interval)

    def send(self, span: TraceSpanObserver) -> None:
        try:
            self.span_queue.put_nowait(span)
        except Exception as e:
            self.logger.warning("Failed adding span to recording queue: %s", e)


class LoggingRecorder(BaseBatchRecorder):
    """Interface for recording spans to the debug log."""

    def __init__(
        self,
        max_queue_size: int = 50000,
        num_workers: int = 5,
        max_span_batch: int = 100,
        batch_wait_interval: float = 0.5,
    ):
        super().__init__(max_queue_size, num_workers, max_span_batch, batch_wait_interval)

    def flush_func(self, spans: List[Dict[str, Any]]) -> None:
        """Write a set of spans to debug log."""
        for span in spans:
            self.logger.debug("Span recording: %s", span)


class NullRecorder(BaseBatchRecorder):
    """Noop recorder."""

    def __init__(
        self,
        max_queue_size: int = 50000,
        num_workers: int = 5,
        max_span_batch: int = 100,
        batch_wait_interval: float = 0.5,
    ):
        super().__init__(max_queue_size, num_workers, max_span_batch, batch_wait_interval)

    def flush_func(self, spans: List[Dict[str, Any]]) -> None:
        return


class RemoteRecorder(BaseBatchRecorder):
    """Interface for recording spans to a remote collector.

    The RemoteRecorder adds spans to an in-memory Queue for a background
    thread worker to process. It currently does not shut down gracefully -
    in the event of parent process exit, any remaining spans will be discarded.
    """

    def __init__(
        self,
        endpoint: str,
        num_conns: int = 5,
        num_workers: int = 5,
        max_queue_size: int = 50000,
        max_span_batch: int = 100,
        batch_wait_interval: float = 0.5,
    ):

        super().__init__(max_queue_size, num_workers, max_span_batch, batch_wait_interval)
        adapter = requests.adapters.HTTPAdapter(pool_connections=num_conns, pool_maxsize=num_conns)
        self.session = requests.Session()
        self.session.mount("http://", adapter)
        self.endpoint = f"http://{endpoint}/api/v1/spans"

    def flush_func(self, spans: List[Dict[str, Any]]) -> None:
        """Send a set of spans to remote collector."""
        try:
            self.session.post(
                self.endpoint,
                data=json.dumps(spans).encode("utf8"),
                headers={"Content-Type": "application/json"},
                timeout=1,
            )
        except RequestException as e:
            self.logger.warning("Error flushing spans: %s", e)


class TraceTooLargeError(Exception):
    pass


class TraceQueueFullError(Exception):
    pass


class SidecarRecorder(Recorder):
    """Interface for recording spans to a POSIX message queue.

    The SidecarRecorder serializes spans to a string representation before
    adding them to the queue.
    """

    def __init__(self, queue_name: str):
        self.queue = MessageQueue(
            "/traces-" + queue_name,
            max_messages=MAX_QUEUE_SIZE,
            max_message_size=MAX_SPAN_SIZE,
        )

    def send(self, span: TraceSpanObserver) -> None:
        # Don't raise exceptions from here. This is called in the
        # request/response path and should finish cleanly.
        serialized_str = json.dumps(span._serialize()).encode("utf8")
        if len(serialized_str) > MAX_SPAN_SIZE:
            logger.warning(
                "Trace too big. Traces published to %s are not allowed to be larger "
                "than %d bytes. Received trace is %d bytes. This can be caused by "
                "an excess amount of tags or a large amount of child spans.",
                self.queue.queue.name,
                MAX_SPAN_SIZE,
                len(serialized_str),
            )
            return
        try:
            self.queue.put(serialized_str, timeout=0)
        except TimedOutError:
            logger.warning(
                "Trace queue %s is full. Is trace sidecar healthy?", self.queue.queue.name
            )


def tracing_client_from_config(
    raw_config: config.RawConfig, log_if_unconfigured: bool = True
) -> TracingClient:
    """Configure and return a tracing client.

    This expects one configuration option and can take many optional ones:

    ``tracing.service_name``
        The name for the service this observer is registered to.
    ``tracing.endpoint`` (optional)
        (Deprecated in favor of the sidecar model.) Destination to record span data.
    ``tracing.queue_name`` (optional)
        Name of POSIX queue where spans are recorded
    ``tracing.max_span_queue_size`` (optional)
        Span processing queue limit.
    ``tracing.num_span_workers`` (optional)
        Number of worker threads for span processing.
    ``tracing.span_batch_interval`` (optional)
        Wait time for span processing in seconds.
    ``tracing.num_conns`` (optional)
        Pool size for remote recorder connection pool.
    ``tracing.sample_rate`` (optional)
        Percentage of unsampled requests to record traces for (e.g. "37%")

    :param raw_config: The application configuration which should have settings
        for the tracing client.
    :param log_if_unconfigured: When the client is not configured, should
        trace spans be logged or discarded silently?
    :return: A configured client.

    """
    cfg = config.parse_config(
        raw_config,
        {
            "tracing": {
                "service_name": config.String,
                "endpoint": config.Optional(config.Endpoint),
                "queue_name": config.Optional(config.String),
                "max_span_queue_size": config.Optional(config.Integer, default=50000),
                "num_span_workers": config.Optional(config.Integer, default=5),
                "span_batch_interval": config.Optional(
                    config.Timespan, default=config.Timespan("500 milliseconds")
                ),
                "num_conns": config.Optional(config.Integer, default=100),
                "sample_rate": config.Optional(
                    config.Fallback(config.Percent, config.Float), default=0.1
                ),
            }
        },
    )

    # pylint: disable=maybe-no-member
    return make_client(
        service_name=cfg.tracing.service_name,
        tracing_endpoint=cfg.tracing.endpoint,
        tracing_queue_name=cfg.tracing.queue_name,
        max_span_queue_size=cfg.tracing.max_span_queue_size,
        num_span_workers=cfg.tracing.num_span_workers,
        span_batch_interval=cfg.tracing.span_batch_interval.total_seconds(),
        num_conns=cfg.tracing.num_conns,
        sample_rate=cfg.tracing.sample_rate,
        log_if_unconfigured=log_if_unconfigured,
    )



import collections
import unittest

from unittest import mock

from baseplate.lib import random


class WeightedChoiceTests(unittest.TestCase):
    def test_empty_list(self):
        with self.assertRaises(ValueError):
            random.WeightedLottery([], lambda i: i)

    def test_negative_weight(self):
        with self.assertRaises(ValueError):
            random.WeightedLottery(["a"], lambda i: -1)

    def test_no_weight(self):
        with self.assertRaises(ValueError):
            random.WeightedLottery(["a"], lambda i: 0)

    @mock.patch("random.random")
    def test_iterator(self, mock_random):
        iterator = (n for n in range(10))

        mock_random.return_value = 0.0
        lottery = random.WeightedLottery(iterator, lambda i: i)
        choice = lottery.pick()
        self.assertEqual(choice, 1)

    @mock.patch("random.random")
    def test_choice(self, mock_random):
        def weight_fn(i):
            if i == "c":
                return 3
            else:
                return 1

        items = ["a", "b", "c", "d"]  # 1  # 1  # 3  # 1

        mock_random.return_value = 0.5
        lottery = random.WeightedLottery(items, weight_fn)
        choice = lottery.pick()
        self.assertEqual(choice, "c")

    def test_distribution(self):
        def weight_fn(i):
            return ord(i) - 96

        items = ["a", "b", "c", "d"]  # 1  # 2  # 3  # 4
        lottery = random.WeightedLottery(items, weight_fn)

        choices = collections.Counter()
        for _ in range(10000):
            choice = lottery.pick()
            choices[choice] += 1

        # we give a bit of fuzz factor here since we're
        # allowing true randomness in this test and don't
        # want spurious failures.
        self.assertLess(abs(1000 - choices["a"]), 150)
        self.assertLess(abs(2000 - choices["b"]), 150)
        self.assertLess(abs(3000 - choices["c"]), 150)
        self.assertLess(abs(4000 - choices["d"]), 150)

    def test_sample_errors(self):
        def weight_fn(i):
            if i == "c":
                return 3
            else:
                return 1

        items = ["a", "b", "c", "d"]  # 1  # 1  # 3  # 1
        lottery = random.WeightedLottery(items, weight_fn)

        with self.assertRaises(ValueError):
            lottery.sample(-1)

        with self.assertRaises(ValueError):
            lottery.sample(5)

    def test_sample(self):
        def weight_fn(i):
            if i == "c":
                return 3
            else:
                return 1

        items = ["a", "b", "c", "d"]  # 1  # 1  # 3  # 1
        lottery = random.WeightedLottery(items, weight_fn)

        for k in range(len(items)):
            samples = lottery.sample(k)
            # we got the right number
            self.assertEqual(len(samples), k)
            # there are no duplicates
            self.assertEqual(len(samples), len(set(samples)))



"""Client library for children of Einhorn."""
import contextlib
import json
import os
import socket


class NotEinhornWorker(Exception):
    pass


def is_worker() -> bool:
    """Return if this process is an Einhorn worker."""
    return os.getppid() == int(os.environ.get("EINHORN_MASTER_PID", -1))


def get_socket_count() -> int:
    """Return how many sockets are bound."""
    if not is_worker():
        raise NotEinhornWorker

    return int(os.environ.get("EINHORN_FD_COUNT", 0))


def get_socket(index: int = 0) -> socket.socket:
    """Get an Einhorn-bound socket from the environment.

    Einhorn can bind multiple sockets (via multiple -b arguments), the
    ``index`` parameter can be used to choose which socket to retrieve.  When
    sockets are bound, Einhorn provides several environment variables to child
    worker processes:

    - EINHORN_FD_COUNT: the number of sockets bound
    - EINHORN_FD_#: for each socket bound, the file descriptor for that socket
    - EINHORN_FD_FAMILY_#: for each socket bound, the protocol family of that
        socket (this is a recent addition, so if it's not present default to
        AF_INET)

    :param index: The socket number to get.

    """
    if not is_worker():
        raise NotEinhornWorker

    fd_count = get_socket_count()
    if not 0 <= index < fd_count:
        raise IndexError

    fileno = int(os.environ[f"EINHORN_FD_{index:d}"])
    family_name = os.environ.get(f"EINHORN_FD_FAMILY_{index:d}", "AF_INET")
    assert family_name.startswith("AF_"), "invalid socket family name"
    family = getattr(socket, family_name)
    return socket.fromfd(fileno, family, socket.SOCK_STREAM)


def ack_startup() -> None:
    """Send acknowledgement that we started up to the Einhorn master."""
    if not is_worker():
        raise NotEinhornWorker

    control_sock_name = os.environ["EINHORN_SOCK_PATH"]
    control_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    control_sock.connect(control_sock_name)

    with contextlib.closing(control_sock):
        control_sock.sendall(
            (json.dumps({"command": "worker:ack", "pid": os.getpid()}) + "\n").encode("utf-8")
        )



import astroid
import pylint.testutils

from baseplate.lint import db_query_string_format_plugin


class TestNoCQLStringFormatChecker(pylint.testutils.CheckerTestCase):
    CHECKER_CLASS = db_query_string_format_plugin.NoDbQueryStringFormatChecker

    def test_finds_variable_binop_string_format_query(self):
        assign_node_a, call_node_b = astroid.extract_node(
            """
        query = "select * from %s" % "users" #@
        cassandra_session.execute(query) #@
        """
        )

        self.checker.visit_assign(assign_node_a)
        self.checker.visit_call(call_node_b)
        self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="database-query-string-format", node=call_node_b)
        )

    def test_finds_variable_call_string_format_query(self):
        assign_node_a, call_node_b = astroid.extract_node(
            """
        query = "select * from %s".format("users") #@
        cassandra_session.execute(query) #@
        """
        )

        self.checker.visit_assign(assign_node_a)
        self.checker.visit_call(call_node_b)
        self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="database-query-string-format", node=call_node_b)
        )

    def test_finds_binop_string_format_query(self):
        call_node_a = astroid.extract_node(
            """
        cassandra_session.execute("select * from %s" % "users") #@
        """
        )

        self.checker.visit_call(call_node_a)
        self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="database-query-string-format", node=call_node_a)
        )

    def test_finds_call_string_format_query(self):
        call_node_a = astroid.extract_node(
            """
        cassandra_session.execute("select * from %s".format("users")) #@
        """
        )

        self.checker.visit_call(call_node_a)
        self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="database-query-string-format", node=call_node_a)
        )

    def test_ignores_variable_non_string_format_query(self):
        assign_node_a, call_node_b = astroid.extract_node(
            """
        query = "select * from users" #@
        cassandra_session.execute(query) #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.visit_call(call_node_b)

    def test_ignores_non_string_format_query(self):
        call_node_a = astroid.extract_node(
            """
        cassandra_session.execute("select * from users") #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_call(call_node_a)

    def test_ignores_no_argument(self):
        call_node_a = astroid.extract_node(
            """
        kb.execute() #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_call(call_node_a)

    def test_variable_reset(self):
        assign_node_a, assign_node_b, call_node_c = astroid.extract_node(
            """
        query = "select * from %s" % "users" #@
        query = "select * from users" #@
        cassandra_session.execute(query) #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.visit_assign(assign_node_b)
            self.checker.visit_call(call_node_c)

    def test_func_variable_reset(self):
        func_node, assign_node_a, assign_node_b, call_node_c = astroid.extract_node(
            """
        def test(cassandra_session): #@
            query = "select * from %s" % "users" #@
            cassandra_session.execute(query)

        def other_test(cassandra_session):
            query = "select * from users" #@
            cassandra_session.execute(query) #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.leave_functiondef(func_node)
            self.checker.visit_assign(assign_node_b)
            self.checker.visit_call(call_node_c)

    def test_class_variable_reset(self):
        class_node, assign_node_a, assign_node_b, call_node_c = astroid.extract_node(
            """
        class Test(): #@
            query = "select * from %s" % "users" #@
            cassandra_session.execute(query)

        query = "select * from users" #@
        cassandra_session.execute(query) #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.leave_classdef(class_node)
            self.checker.visit_assign(assign_node_b)
            self.checker.visit_call(call_node_c)

    def test_module_variable_reset(self):
        mod_node, assign_node_a, assign_node_b, call_node_c = astroid.extract_node(
            """
        import test #@
        query = "select * from %s" % "users" #@
        cassandra_session.execute(query)

        import other_test
        query = "select * from users" #@
        cassandra_session.execute(query) #@
        """
        )

        with self.assertNoMessages():
            self.checker.visit_assign(assign_node_a)
            self.checker.leave_module(mod_node)
            self.checker.visit_assign(assign_node_b)
            self.checker.visit_call(call_node_c)



import time
import unittest

from unittest import mock

try:
    from cassandra import InvalidRequest, ConsistencyLevel
    from cassandra.cluster import ExecutionProfile
    from cassandra.concurrent import execute_concurrent_with_args
    from cassandra.query import dict_factory, named_tuple_factory
except ImportError:
    raise unittest.SkipTest("cassandra-driver is not installed")

from baseplate.clients.cassandra import CassandraClient
from baseplate import Baseplate

from . import TestBaseplateObserver, get_endpoint_or_skip_container


cassandra_endpoint = get_endpoint_or_skip_container("cassandra", 9042)


class CassandraTests(unittest.TestCase):
    def setUp(self):
        self.baseplate_observer = TestBaseplateObserver()

        profiles = {"foo": ExecutionProfile(consistency_level=ConsistencyLevel.QUORUM)}

        baseplate = Baseplate(
            {
                "cassandra.contact_points": cassandra_endpoint.address.host,
                "cassandra_no_prof.contact_points": cassandra_endpoint.address.host,
            }
        )
        baseplate.register(self.baseplate_observer)
        baseplate.configure_context(
            {
                "cassandra_no_prof": CassandraClient(keyspace="system"),
                "cassandra": CassandraClient(keyspace="system", execution_profiles=profiles),
            }
        )

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple_query(self):
        with self.server_span:
            self.context.cassandra.execute("SELECT * FROM system.local;")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)
        span_observer.assert_tag("statement", "SELECT * FROM system.local;")

    def test_error_in_query(self):
        with self.server_span:
            with self.assertRaises(InvalidRequest):
                self.context.cassandra.execute("SELECT * FROM does_not_exist;")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNotNone(span_observer.on_finish_exc_info)

    def test_async(self):
        with self.server_span:
            future = self.context.cassandra.execute_async("SELECT * FROM system.local;")
            future.result()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)
        span_observer.assert_tag("statement", "SELECT * FROM system.local;")

    def test_properties(self):
        with self.server_span:
            self.assertIsNotNone(self.context.cassandra_no_prof.cluster)
            self.assertIsNotNone(self.context.cassandra_no_prof.encoder)
            self.assertEqual(self.context.cassandra_no_prof.keyspace, "system")

            self.assertEqual(self.context.cassandra_no_prof.row_factory, named_tuple_factory)
            self.context.cassandra_no_prof.row_factory = dict_factory
            self.assertEqual(self.context.cassandra_no_prof.row_factory, dict_factory)

    def test_prepared_statements(self):
        with self.server_span:
            statement = self.context.cassandra.prepare("SELECT * FROM system.local;")
            self.context.cassandra.execute(statement)

    def test_async_callback_fail(self):
        # mock threading.Event so that Event.wait() returns immediately
        event = mock.patch("baseplate.clients.cassandra.Event")
        event.start()
        self.addCleanup(event.stop)

        # mock the on_execute_complete callback to be slow
        def on_execute_complete(result, span, event):
            time.sleep(0.005)
            span.finish()
            event.set()

        on_execute_complete = mock.patch(
            "baseplate.clients.cassandra._on_execute_complete", side_effect=on_execute_complete
        )
        on_execute_complete.start()
        self.addCleanup(on_execute_complete.stop)

        with self.server_span:
            future = self.context.cassandra.execute_async("SELECT * FROM system.local;")
            future.result()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertFalse(span_observer.on_finish_called)

    def test_async_callback_pass(self):
        # mock the on_execute_complete callback to be slow
        def on_execute_complete(result, span, event):
            time.sleep(0.005)
            span.finish()
            event.set()

        on_execute_complete = mock.patch(
            "baseplate.clients.cassandra._on_execute_complete", side_effect=on_execute_complete
        )
        on_execute_complete.start()
        self.addCleanup(on_execute_complete.stop)

        with self.server_span:
            future = self.context.cassandra.execute_async("SELECT * FROM system.local;")
            future.result()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_finish_called)

    def test_async_callback_too_slow(self):
        # mock the on_execute_complete callback to be slow
        def on_execute_complete(result, span, event):
            time.sleep(0.02)
            span.finish()
            event.set()

        on_execute_complete = mock.patch(
            "baseplate.clients.cassandra._on_execute_complete", side_effect=on_execute_complete
        )
        on_execute_complete.start()
        self.addCleanup(on_execute_complete.stop)

        with self.server_span:
            future = self.context.cassandra.execute_async("SELECT * FROM system.local;")
            future.result()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertFalse(span_observer.on_finish_called)


class CassandraConcurrentTests(unittest.TestCase):
    def setUp(self):
        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate({"cassandra.contact_points": cassandra_endpoint.address.host})
        baseplate.register(self.baseplate_observer)
        baseplate.configure_context({"cassandra": CassandraClient(keyspace="system")})

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_execute_concurrent_with_args(self):
        with self.server_span:
            statement = self.context.cassandra.prepare('SELECT * FROM system.local WHERE "key"=?')
            params = [(_key,) for _key in ["local", "other"]]
            results = execute_concurrent_with_args(self.context.cassandra, statement, params)

        server_span_observer = self.baseplate_observer.get_only_child()
        self.assertEqual(len(server_span_observer.children), 3)
        for span_observer in server_span_observer.children:
            self.assertTrue(span_observer.on_start_called)
            self.assertTrue(span_observer.on_finish_called)
            self.assertIsNone(span_observer.on_finish_exc_info)
            span_observer.assert_tag("statement", 'SELECT * FROM system.local WHERE "key"=?')

        self.assertEqual(len(results), 2)
        self.assertTrue(results[0].success)
        self.assertTrue(results[1].success)



import unittest

from unittest import mock

from pymemcache.client.base import PooledClient
from redis import StrictRedis

from baseplate.lib import ratelimit
from baseplate.lib.ratelimit.backends import RateLimitBackend
from baseplate.lib.ratelimit.backends.memcache import MemcacheRateLimitBackend
from baseplate.lib.ratelimit.backends.redis import RedisRateLimitBackend


class MockRateLimitBackend(RateLimitBackend):
    def __init__(self):
        self.counter = 0

    def consume(self, key, amount, max, bucket_size):
        self.counter += amount
        return self.counter <= max


class RateLimiterTests(unittest.TestCase):
    def setUp(self):
        ratelimit_backend = MockRateLimitBackend()
        self.ratelimiter = ratelimit.RateLimiter(ratelimit_backend, 10, 60)

    def test_consume(self):
        self.ratelimiter.consume("user_12345")

    def test_consume_over_allowance(self):
        with self.assertRaises(ratelimit.RateLimitExceededException):
            self.ratelimiter.consume("user_12345", amount=11)


class RedisRateLimitBackendTest(unittest.TestCase):
    def setUp(self):
        self.amount = 10
        redis = mock.create_autospec(StrictRedis)
        pipeline_context = redis.pipeline.return_value.__enter__.return_value
        pipeline_context.execute.return_value = [self.amount]
        self.ratelimit_backend = RedisRateLimitBackend(redis)

    def test_consume(self):
        self.assertTrue(self.ratelimit_backend.consume("user_12345", self.amount, 10, 60))

    def test_consume_over_max(self):
        self.assertFalse(self.ratelimit_backend.consume("user_12345", self.amount, 5, 60))


class MemcacheRateLimitBackendTest(unittest.TestCase):
    def setUp(self):
        self.amount = 10
        memcache = mock.create_autospec(PooledClient)
        memcache.incr.return_value = self.amount
        self.ratelimit_backend = MemcacheRateLimitBackend(memcache)

    def test_consume(self):
        self.assertTrue(self.ratelimit_backend.consume("user_12345", self.amount, 10, 60))

    def test_consume_over_max(self):
        self.assertFalse(self.ratelimit_backend.consume("user_12345", self.amount, 5, 60))



import unittest

from unittest import mock

from prometheus_client import REGISTRY

try:
    import cassandra

    del cassandra
except ImportError:
    raise unittest.SkipTest("cassandra-driver is not installed")

import baseplate
import logging
from baseplate.lib.config import ConfigurationError
from baseplate.clients.cassandra import (
    cluster_from_config,
    CassandraCallbackArgs,
    CassandraPrometheusLabels,
    CassandraSessionAdapter,
    REQUEST_TIME,
    REQUEST_ACTIVE,
    REQUEST_TOTAL,
    _on_execute_complete,
    _on_execute_failed,
)
from baseplate.lib.secrets import SecretsStore

logger = logging.getLogger(__name__)


class ClusterFromConfigTests(unittest.TestCase):
    def test_empty_config(self):
        with self.assertRaises(ConfigurationError):
            cluster_from_config({})

    def test_contact_points(self):
        cluster = cluster_from_config({"cassandra.contact_points": "127.0.1.1, 127.0.1.2"})

        self.assertEqual(cluster.contact_points, ["127.0.1.1", "127.0.1.2"])

    def test_port(self):
        cluster = cluster_from_config(
            {"cassandra.contact_points": "127.0.1.1", "cassandra.port": "9999"}
        )

        self.assertEqual(cluster.port, 9999)

    def test_kwarg_passthrough(self):
        cluster = cluster_from_config({"cassandra.contact_points": "127.0.0.1"}, protocol_version=3)

        self.assertEqual(cluster.protocol_version, 3)

    def test_alternate_prefix(self):
        cluster = cluster_from_config(
            {"noodle.contact_points": "127.0.1.1, 127.0.1.2"}, prefix="noodle."
        )

        self.assertEqual(cluster.contact_points, ["127.0.1.1", "127.0.1.2"])

    def test_credentials_but_no_secrets(self):
        with self.assertRaises(TypeError):
            cluster_from_config(
                {
                    "cassandra.contact_points": "127.0.0.1",
                    "cassandra.credentials_secret": "secret/foo/bar",
                }
            )

    def test_credentials(self):
        secrets = mock.Mock(autospec=SecretsStore)
        cluster = cluster_from_config(
            {
                "cassandra.contact_points": "127.0.0.1",
                "cassandra.credentials_secret": "secret/foo/bar",
            },
            secrets=secrets,
        )
        self.assertIsNotNone(cluster.auth_provider)


class CassandraSessionAdapterTests(unittest.TestCase):
    def setUp(self):
        # cleaning up prom registry
        REQUEST_TIME.clear()
        REQUEST_ACTIVE.clear()
        REQUEST_TOTAL.clear()

        self.session = mock.MagicMock()
        self.prepared_statements = {}
        self.mock_server_span = mock.MagicMock(spec=baseplate.ServerSpan)
        self.adapter = CassandraSessionAdapter(
            "test", self.mock_server_span, self.session, self.prepared_statements
        )

    def test_prepare(self):
        statement = "SELECT foo from bar;"
        self.adapter.prepare(statement, cache=True)
        self.adapter.prepare(statement, cache=True)
        # Assert that when preparing an identical statement twice, the CQL
        # session only gets one prepare.
        self.session.prepare.assert_called_once_with(statement)
        self.adapter.prepare(statement, cache=False)
        # Assert that preparing an identical statement with cache=False always
        # prepares.
        calls = [mock.call(statement), mock.call(statement)]
        self.session.prepare.assert_has_calls(calls)

    def test_execute_async_prom_metrics(self):
        self.session.keyspace = "keyspace"  # mocking keyspace name
        self.adapter.execute_async("SELECT foo from bar;")

        self.assertEqual(
            REGISTRY.get_sample_value(
                "cassandra_client_active_requests",
                {
                    "cassandra_client_name": "test",  # client name defaults to name when not provided
                    "cassandra_keyspace": "keyspace",
                    "cassandra_query_name": "",
                },
            ),
            1,
        )

    def test_execute_async_with_query_name_prom_metrics(self):
        self.session.keyspace = "keyspace"  # mocking keyspace name
        self.adapter.execute_async("SELECT foo from bar;", query_name="foo_bar")

        self.assertEqual(
            REGISTRY.get_sample_value(
                "cassandra_client_active_requests",
                {
                    "cassandra_client_name": "test",
                    "cassandra_keyspace": "keyspace",
                    "cassandra_query_name": "foo_bar",
                },
            ),
            1,
        )

    def test_execute_async_prom_metrics_client_name_specified(self):
        self.adapter = CassandraSessionAdapter(
            "test",
            self.mock_server_span,
            self.session,
            self.prepared_statements,
            prometheus_client_name="test_client_name",
        )
        self.session.keyspace = "keyspace"  # mocking keyspace name
        self.adapter.execute_async("SELECT foo from bar;")

        self.assertEqual(
            REGISTRY.get_sample_value(
                "cassandra_client_active_requests",
                {
                    "cassandra_client_name": "test_client_name",
                    "cassandra_keyspace": "keyspace",
                    "cassandra_query_name": "",
                },
            ),
            1,
        )

    def test_execute_async_prom_metrics_client_name_empty(self):
        self.adapter = CassandraSessionAdapter(
            "test",
            self.mock_server_span,
            self.session,
            self.prepared_statements,
            prometheus_client_name="",
        )
        self.session.keyspace = "keyspace"  # mocking keyspace name
        self.adapter.execute_async("SELECT foo from bar;")

        self.assertEqual(
            REGISTRY.get_sample_value(
                "cassandra_client_active_requests",
                {
                    "cassandra_client_name": "",
                    "cassandra_keyspace": "keyspace",
                    "cassandra_query_name": "",
                },
            ),
            1,
        )


class CassandraTests(unittest.TestCase):
    def setUp(self):
        REQUEST_TIME.clear()
        REQUEST_ACTIVE.clear()
        REQUEST_TOTAL.clear()

    def test_prom__on_execute_complete(self):
        result = mock.MagicMock()
        span = mock.MagicMock()
        event = mock.MagicMock()
        start_time = 1.0

        prom_labels_tuple = CassandraPrometheusLabels(
            cassandra_client_name="test_client_name",
            cassandra_keyspace="keyspace",
            cassandra_query_name="",
        )

        _on_execute_complete(
            result,
            CassandraCallbackArgs(
                span=span,
                start_time=start_time,
                prom_labels=prom_labels_tuple,
            ),
            event,
        )
        prom_labels = prom_labels_tuple._asdict()
        prom_labels_w_success = {**prom_labels, **{"cassandra_success": "true"}}

        self.assertEquals(
            REGISTRY.get_sample_value("cassandra_client_requests_total", prom_labels_w_success), 1
        )

        # we start from 0 here since this is a unit test, so -1 is the expected result
        self.assertEquals(
            REGISTRY.get_sample_value("cassandra_client_active_requests", prom_labels), -1
        )

        self.assertEquals(
            REGISTRY.get_sample_value(
                "cassandra_client_latency_seconds_bucket",
                {**prom_labels_w_success, **{"le": "+Inf"}},
            ),
            1,
        )

    def test_prom__on_execute_failed(self):
        result = mock.MagicMock()
        span = mock.MagicMock()
        event = mock.MagicMock()
        start_time = 1.0

        prom_labels_tuple = CassandraPrometheusLabels(
            cassandra_client_name="test_client_name",
            cassandra_keyspace="keyspace",
            cassandra_query_name="",
        )

        _on_execute_failed(
            result,
            CassandraCallbackArgs(
                span=span,
                start_time=start_time,
                prom_labels=prom_labels_tuple,
            ),
            event,
        )
        prom_labels = prom_labels_tuple._asdict()
        prom_labels_w_success = {**prom_labels, **{"cassandra_success": "false"}}

        self.assertEquals(
            REGISTRY.get_sample_value("cassandra_client_requests_total", prom_labels_w_success), 1
        )

        # we start from 0 here since this is a unit test, so -1 is the expected result
        self.assertEquals(
            REGISTRY.get_sample_value("cassandra_client_active_requests", prom_labels), -1
        )

        self.assertEquals(
            REGISTRY.get_sample_value(
                "cassandra_client_latency_seconds_bucket",
                {**prom_labels_w_success, **{"le": "+Inf"}},
            ),
            1,
        )



import unittest

from unittest import mock

from baseplate.lib.events import EventQueue
from baseplate.lib.events import EventQueueFullError
from baseplate.lib.events import EventTooLargeError
from baseplate.lib.events import MAX_EVENT_SIZE
from baseplate.lib.message_queue import MessageQueue
from baseplate.lib.message_queue import TimedOutError


class EventQueueTests(unittest.TestCase):
    @mock.patch("baseplate.lib.events.MessageQueue", autospec=MessageQueue)
    def setUp(self, MessageQueue):
        self.message_queue = MessageQueue.return_value
        self.mock_serializer = mock.Mock()
        self.queue = EventQueue("test", event_serializer=self.mock_serializer)

    def test_send_event(self):
        self.mock_serializer.return_value = "i_am_serialized"
        event = object()

        self.queue.put(event)

        self.assertEqual(self.message_queue.put.call_count, 1)
        self.mock_serializer.assert_called_with(event)
        args, kwargs = self.message_queue.put.call_args
        self.assertEqual(args[0], self.mock_serializer.return_value)

    def test_event_too_large(self):
        self.mock_serializer.return_value = "x" * (MAX_EVENT_SIZE + 1)

        with self.assertRaises(EventTooLargeError):
            self.queue.put(object())

    def test_event_queue_full(self):
        self.mock_serializer.return_value = ""

        self.message_queue.put.side_effect = TimedOutError

        with self.assertRaises(EventQueueFullError):
            self.queue.put(object())



from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib.ratelimit.backends import RateLimitBackend


class RateLimitExceededException(Exception):
    """This exception gets raised whenever a rate limit is exceeded."""


class RateLimiterContextFactory(ContextFactory):
    """RateLimiter context factory.

    :param backend_factory: An instance of
        :py:class:`baseplate.clients.ContextFactory`. The context factory must
        return an instance of :py:class:`baseplate.lib.ratelimit.backends.RateLimitBackend`
    :param allowance: The maximum allowance allowed per key.
    :param interval: The interval (in seconds) to reset allowances.

    """

    def __init__(self, backend_factory: ContextFactory, allowance: int, interval: int):
        if allowance < 1:
            raise ValueError("minimum allowance is 1")
        if interval < 1:
            raise ValueError("minimum interval is 1")
        if not isinstance(backend_factory, ContextFactory):
            raise TypeError("backend_factory must be an instance of ContextFactory")

        self.backend_factory = backend_factory
        self.allowance = allowance
        self.interval = interval

    def make_object_for_context(self, name: str, span: Span) -> "RateLimiter":
        backend = self.backend_factory.make_object_for_context(name, span)
        return RateLimiter(backend, self.allowance, self.interval)


class RateLimiter:
    """A class for rate limiting actions.

    :param backend: The backend to use for storing rate limit counters.
    :param allowance: The maximum allowance allowed per key.
    :param interval: The interval (in seconds) to reset allowances.

    """

    def __init__(self, backend: RateLimitBackend, allowance: int, interval: int):
        if allowance < 1:
            raise ValueError("minimum allowance is 1")
        if interval < 1:
            raise ValueError("minimum interval is 1")
        if not isinstance(backend, RateLimitBackend):
            raise TypeError("backend must be an instance of RateLimitBackend")

        self.backend = backend
        self.allowance = allowance
        self.interval = interval

    def consume(self, key: str, amount: int = 1) -> None:
        """Consume the given `amount` from the allowance for the given `key`.

        This will raise
        :py:class:`baseplate.lib.ratelimit.RateLimitExceededException` if the
        allowance for `key` is exhausted.

        :param key: The name of the rate limit bucket to consume from.
        :param amount: The amount to consume from the rate limit bucket.

        """
        if not self.backend.consume(key, amount, self.allowance, self.interval):
            raise RateLimitExceededException("Rate limit exceeded.")



"""Prometheus exporter server.

This is automatically run by `baseplate-serve` whenever the `prometheus-client`
package is installed.

This exporter is designed for multi-process usage, like under Einhorn.  The
PROMETHEUS_MULTIPROC_DIR environment variable must be set to the path of an
existing writeable directory where individual application workers will write
their metrics. Each application worker will be serving the exporter as well and
can aggregate and serve metrics for all workers.

"""
import atexit
import logging
import os
import sys

from typing import Iterable
from typing import TYPE_CHECKING

from gevent.pywsgi import LoggingLogAdapter
from gevent.pywsgi import WSGIServer
from prometheus_client import CollectorRegistry
from prometheus_client import CONTENT_TYPE_LATEST
from prometheus_client import generate_latest
from prometheus_client import multiprocess

from baseplate.lib.config import Endpoint
from baseplate.lib.config import EndpointConfiguration
from baseplate.server.net import bind_socket


if TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse  # pylint: disable=import-error,no-name-in-module
    from _typeshed.wsgi import WSGIEnvironment  # pylint: disable=import-error,no-name-in-module


logger = logging.getLogger(__name__)
PROMETHEUS_EXPORTER_ADDRESS = Endpoint("0.0.0.0:6060")
METRICS_ENDPOINT = "/metrics"


def export_metrics(environ: "WSGIEnvironment", start_response: "StartResponse") -> Iterable[bytes]:
    if environ["PATH_INFO"] != METRICS_ENDPOINT:
        start_response("404 Not Found", [("Content-Type", "text/plain")])
        return [b"Not Found"]

    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    data = generate_latest(registry)
    response_headers = [("Content-type", CONTENT_TYPE_LATEST), ("Content-Length", str(len(data)))]
    start_response("200 OK", response_headers)
    return [data]


def start_prometheus_exporter(address: EndpointConfiguration = PROMETHEUS_EXPORTER_ADDRESS) -> None:
    if "PROMETHEUS_MULTIPROC_DIR" not in os.environ:
        logger.error(
            "prometheus-client is installed but PROMETHEUS_MULTIPROC_DIR is not set to a writeable directory."
        )
        sys.exit(1)

    atexit.register(multiprocess.mark_process_dead, os.getpid())

    server_socket = bind_socket(address)
    server = WSGIServer(
        server_socket,
        application=export_metrics,
        log=LoggingLogAdapter(logger, level=logging.DEBUG),
        error_log=LoggingLogAdapter(logger, level=logging.ERROR),
    )
    logger.info(
        "Prometheus metrics exported on server listening on %s%s",
        address,
        METRICS_ENDPOINT,
    )
    server.start()


def start_prometheus_exporter_for_sidecar() -> None:
    port = os.environ.get("BASEPLATE_SIDECAR_ADMIN_PORT")
    if port is None:
        logger.error(
            "BASEPLATE_SIDECAR_ADMIN_PORT must be set for sidecar to expose Prometheus metrics."
        )
    else:
        endpoint = Endpoint("0.0.0.0:" + port)
        start_prometheus_exporter(endpoint)



#
# Autogenerated by Thrift Compiler (0.14.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:slots
#
import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.Thrift import TFrozenDict
from thrift.Thrift import TMessageType
from thrift.Thrift import TType
from thrift.transport import TTransport
from thrift.TRecursive import fix_spec

import baseplate.thrift.ttypes

all_structs = []


class ExpectedException(TException):

    __slots__ = ()

    def __setattr__(self, *args):
        raise TypeError("can't modify immutable instance")

    def __delattr__(self, *args):
        raise TypeError("can't modify immutable instance")

    def __hash__(self):
        return hash(self.__class__) ^ hash(())

    @classmethod
    def read(cls, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and cls.thrift_spec is not None
        ):
            return iprot._fast_decode(None, iprot, [cls, cls.thrift_spec])
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()
        return cls()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("ExpectedException")
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


class ExampleStruct(object):
    """
    Attributes:
     - string_field
     - int_field

    """

    __slots__ = (
        "string_field",
        "int_field",
    )

    def __init__(
        self,
        string_field=None,
        int_field=None,
    ):
        self.string_field = string_field
        self.int_field = int_field

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.string_field = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I64:
                    self.int_field = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("ExampleStruct")
        if self.string_field is not None:
            oprot.writeFieldBegin("string_field", TType.STRING, 1)
            oprot.writeString(
                self.string_field.encode("utf-8") if sys.version_info[0] == 2 else self.string_field
            )
            oprot.writeFieldEnd()
        if self.int_field is not None:
            oprot.writeFieldBegin("int_field", TType.I64, 2)
            oprot.writeI64(self.int_field)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


all_structs.append(ExpectedException)
ExpectedException.thrift_spec = ()
all_structs.append(ExampleStruct)
ExampleStruct.thrift_spec = (
    None,  # 0
    (
        1,
        TType.STRING,
        "string_field",
        "UTF8",
        None,
    ),  # 1
    (
        2,
        TType.I64,
        "int_field",
        None,
        None,
    ),  # 2
)
fix_spec(all_structs)
del all_structs



import argparse
import configparser
import logging
import urllib.parse

from typing import Optional

import requests

from baseplate import __version__ as baseplate_version
from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.lib.message_queue import MessageQueue
from baseplate.lib.message_queue import TimedOutError
from baseplate.lib.metrics import metrics_client_from_config
from baseplate.lib.retry import RetryPolicy
from baseplate.observers.tracing import MAX_QUEUE_SIZE
from baseplate.observers.tracing import MAX_SPAN_SIZE
from baseplate.server import EnvironmentInterpolation
from baseplate.sidecars import BatchFull
from baseplate.sidecars import RawJSONBatch
from baseplate.sidecars import SerializedBatch
from baseplate.sidecars import TimeLimitedBatch


logger = logging.getLogger(__name__)

# seconds to wait for response from Zipkin server
POST_TIMEOUT_DEFAULT = 3

# maximum size (in bytes) of a batch of traces
MAX_BATCH_SIZE_DEFAULT = 500 * 1024
# maximum time (seconds) a traces can sit around while we wait for more
# messages to batch
MAX_BATCH_AGE = 1

# maximum number of retries when publishing traces
RETRY_LIMIT_DEFAULT = 10


class MaxRetriesError(Exception):
    pass


class TraceBatch(RawJSONBatch):
    def __init__(self, max_size: int = MAX_BATCH_SIZE_DEFAULT):
        super().__init__(max_size)


class ZipkinPublisher:
    """Zipkin trace publisher."""

    def __init__(
        self,
        zipkin_api_url: str,
        metrics_client: metrics.Client,
        post_timeout: int = POST_TIMEOUT_DEFAULT,
        retry_limit: int = RETRY_LIMIT_DEFAULT,
        num_conns: int = 5,
    ):

        adapter = requests.adapters.HTTPAdapter(pool_connections=num_conns, pool_maxsize=num_conns)
        parsed_url = urllib.parse.urlparse(zipkin_api_url)
        self.session = requests.Session()
        self.session.headers[
            "User-Agent"
        ] = f"baseplate.py-{self.__class__.__name__}/{baseplate_version}"
        self.session.mount(f"{parsed_url.scheme}://", adapter)
        self.endpoint = f"{zipkin_api_url}/spans"
        self.metrics = metrics_client
        self.post_timeout = post_timeout
        self.retry_limit = retry_limit

    def publish(self, payload: SerializedBatch) -> None:
        """Publish spans to Zipkin API.

        :param payload: Count and payload to publish.
        """
        if not payload.item_count:
            return

        logger.info("Sending batch of %d traces", payload.item_count)
        headers = {
            "User-Agent": "baseplate-trace-publisher/1.0",
            "Content-Type": "application/json",
        }
        for _ in RetryPolicy.new(attempts=self.retry_limit):
            try:
                with self.metrics.timer("post"):
                    response = self.session.post(
                        self.endpoint,
                        data=payload.serialized,
                        headers=headers,
                        timeout=self.post_timeout,
                        stream=False,
                    )
                response.raise_for_status()
            except requests.HTTPError as exc:
                self.metrics.counter("error.http").increment()
                response = getattr(exc, "response", None)
                if response is not None:
                    logger.exception("HTTP Request failed. Error: %s", response.text)
                    # If client error, skip retries
                    if response.status_code < 500:
                        self.metrics.counter("error.http.client").increment()
                        return
                else:
                    logger.exception("HTTP Request failed. Response not available")
            except OSError:
                self.metrics.counter("error.io").increment()
                logger.exception("HTTP Request failed")
            else:
                self.metrics.counter("sent").increment(payload.item_count)
                return

        raise MaxRetriesError(
            f"ZipkinPublisher exhausted allowance of {self.retry_limit:d} retries."
        )


def publish_traces() -> None:
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    arg_parser.add_argument(
        "--queue-name",
        default="main",
        help="name of trace queue / publisher config (default: main)",
    )
    arg_parser.add_argument(
        "--debug", default=False, action="store_true", help="enable debug logging"
    )
    arg_parser.add_argument(
        "--app-name",
        default="main",
        metavar="NAME",
        help="name of app to load from config_file (default: main)",
    )
    args = arg_parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING
    logging.basicConfig(level=level)

    config_parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    config_parser.read_file(args.config_file)

    publisher_raw_cfg = dict(config_parser.items("trace-publisher:" + args.queue_name))
    publisher_cfg = config.parse_config(
        publisher_raw_cfg,
        {
            "zipkin_api_url": config.DefaultFromEnv(config.Endpoint, "BASEPLATE_ZIPKIN_API_URL"),
            "post_timeout": config.Optional(config.Integer, POST_TIMEOUT_DEFAULT),
            "max_batch_size": config.Optional(config.Integer, MAX_BATCH_SIZE_DEFAULT),
            "retry_limit": config.Optional(config.Integer, RETRY_LIMIT_DEFAULT),
            "max_queue_size": config.Optional(config.Integer, MAX_QUEUE_SIZE),
        },
    )

    trace_queue = MessageQueue(
        "/traces-" + args.queue_name,
        max_messages=publisher_cfg.max_queue_size,
        max_message_size=MAX_SPAN_SIZE,
    )

    # pylint: disable=maybe-no-member
    inner_batch = TraceBatch(max_size=publisher_cfg.max_batch_size)
    batcher = TimeLimitedBatch(inner_batch, MAX_BATCH_AGE)
    metrics_client = metrics_client_from_config(publisher_raw_cfg)
    publisher = ZipkinPublisher(
        publisher_cfg.zipkin_api_url.address,
        metrics_client,
        post_timeout=publisher_cfg.post_timeout,
    )

    while True:
        message: Optional[bytes]

        try:
            message = trace_queue.get(timeout=0.2)
        except TimedOutError:
            message = None

        try:
            batcher.add(message)
        except BatchFull:
            serialized = batcher.serialize()
            publisher.publish(serialized)
            batcher.reset()
            batcher.add(message)


if __name__ == "__main__":
    publish_traces()



# pylint: disable=invalid-name
"""Configuration parsing and validation.

This module provides ``parse_config`` which turns a dictionary of stringy keys
and values into a structured and typed configuration object.

For example, an INI file like the following:

.. highlight:: ini

.. include:: ../../../config_example.ini
   :literal:

Might be parsed like the following. Note: when running under the baseplate
server, The ``config_parser.items(...)`` step is taken care of for you and
``raw_config`` is passed as the only argument to your factory function.

.. highlight:: py

.. testsetup:: overview

    import os
    import configparser
    from baseplate import config
    from tempfile import NamedTemporaryFile
    config_parser = configparser.RawConfigParser()
    config_parser.read_file(open("docs/config_example.ini"))

    tempfile = NamedTemporaryFile()
    tempfile.write(b"cool")
    tempfile.flush()
    config_parser.set("app:main", "some_file", tempfile.name)

    env_var_name = "BASEPLATE_DEFAULT_VALUE"
    os.environ[env_var_name] = "default"

.. doctest:: overview

    >>> raw_config = dict(config_parser.items("app:main"))

    >>> CARDS = config.OneOf(clubs=1, spades=2, diamonds=3, hearts=4)
    >>> cfg = config.parse_config(raw_config, {
    ...     "simple": config.Boolean,
    ...     "cards": config.TupleOf(CARDS),
    ...     "nested": {
    ...         "once": config.Integer,
    ...
    ...         "really": {
    ...             "deep": config.Timespan,
    ...         },
    ...     },
    ...     "some_file": config.File(mode="r"),
    ...     "optional": config.Optional(config.Integer, default=9001),
    ...     "sample_rate": config.Percent,
    ...     "interval": config.Fallback(config.Timespan, config.Integer),
    ...     "default_from_env": config.DefaultFromEnv(config.String, env_var_name),
    ... })

    >>> print(cfg.simple)
    True

    >>> print(cfg.cards)
    [1, 2, 3]

    >>> print(cfg.nested.really.deep)
    0:00:03

    >>> cfg.some_file.read()
    'cool'

    >>> cfg.some_file.close()

    >>> cfg.sample_rate
    0.371

    >>> print(cfg.interval)
    0:00:30

    >>> print(cfg.default_from_env)
    blah

.. testcleanup:: overview

    tempfile.close()

"""
import base64
import datetime
import functools
import grp
import os
import pwd
import re
import socket

from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import IO
from typing import NamedTuple
from typing import NewType
from typing import Optional as OptionalType
from typing import Sequence
from typing import Set
from typing import TypeVar
from typing import Union


class ConfigurationError(Exception):
    """Raised when the configuration violates the spec."""

    def __init__(self, key: str, error: Union[str, Exception]):
        super().__init__(f"{key}: {error}")
        self.key = key
        self.error = error


def String(text: str) -> str:  # noqa: D401
    """A raw string."""
    if not text:
        raise ValueError("no value specified")
    return text


def Float(text: str) -> float:  # noqa: D401
    """A floating-point number."""
    return float(text)


def Integer(
    text: OptionalType[str] = None, base: int = 10
) -> Union[int, Callable[[str], int]]:  # noqa: D401
    """An integer.

    To prevent mistakes, this will raise an error if the user attempts
    to configure a non-whole number.

    :param base: (Optional) If specified, the base of the integer to parse.

    """
    if text is not None:
        # this allows the original config.Integer format
        return int(text, base=base)

    # and this allows the base to be specified as config.Integer(base=N)
    return functools.partial(int, base=base)


def Boolean(text: str) -> bool:  # noqa: D401
    """True or False, case insensitive."""
    parser = OneOf(true=True, false=False)
    return parser(text.lower())


class InternetAddress(NamedTuple):
    host: str
    port: int

    def __str__(self) -> str:
        return f"{self.host}:{self.port}"


class EndpointConfiguration(NamedTuple):
    """A description of a remote endpoint.

    This is a 2-tuple of (``family`` and ``address``).

    ``family``
        One of :py:data:`socket.AF_INET` or :py:data:`socket.AF_UNIX`.

    ``address``
        An address appropriate for the ``family``.

    .. seealso:: :py:func:`baseplate.lib.config.Endpoint`

    """

    family: socket.AddressFamily  # pylint: disable=no-member
    address: Union[InternetAddress, str]

    def __str__(self) -> str:
        return str(self.address)


def Endpoint(text: str) -> EndpointConfiguration:  # noqa: D401
    """A remote endpoint to connect to.

    Returns an :py:class:`EndpointConfiguration`.

    If the endpoint is a hostname:port pair, the ``family`` will be
    :py:data:`socket.AF_INET` and ``address`` will be a two-tuple of host and
    port, as expected by :py:mod:`socket`.

    If the endpoint contains a slash (``/``), it will be interpreted as a path
    to a UNIX domain socket. The ``family`` will be :py:data:`socket.AF_UNIX`
    and ``address`` will be the path as a string.

    """
    if not text:
        raise ValueError("no endpoint specified")

    if "/" in text:
        return EndpointConfiguration(socket.AF_UNIX, text)

    host, sep, port = text.partition(":")
    if sep != ":":
        raise ValueError("no port specified")
    address = InternetAddress(host, int(port))
    return EndpointConfiguration(socket.AF_INET, address)


def Base64(text: str) -> bytes:  # noqa: D401
    """A base64 encoded block of data.

    This is useful for arbitrary binary blobs.

    """
    if not text:
        raise ValueError("expected base64 encoded data")

    try:
        return base64.b64decode(text)
    except TypeError as exc:
        raise ValueError(*exc.args)


def File(mode: str = "r") -> Callable[[str], IO]:  # noqa: D401
    """A path to a file.

    This takes a path to a file and returns an open file object, like
    returned by :py:func:`open`.

    :param mode: an optional string that specifies the mode in which the file
        is opened.

    """

    def open_file(text: str) -> IO:
        try:
            return open(text, mode=mode, encoding="UTF-8")  # pylint: disable=R1732
        except OSError:
            raise ValueError(f"could not open file: {text}")

    return open_file


def Timespan(text: str) -> datetime.timedelta:  # noqa: D401
    """A span of time.

    This takes a string of the form "1 second" or "3 days" and returns a
    :py:class:`datetime.timedelta` representing that span of time.

    Units supported are: milliseconds, seconds, minutes, hours, days.

    """
    scale_by_unit = {
        "millisecond": 0.001,
        "second": 1,
        "minute": 60,
        "hour": 60 * 60,
        "day": 24 * 60 * 60,
    }

    parts = text.split()
    if len(parts) != 2:
        raise ValueError("invalid specification")
    count_text, unit = parts

    count = int(count_text)
    unit = unit.rstrip("s")  # depluralize

    try:
        scale = scale_by_unit[unit]
    except KeyError:
        raise ValueError("unknown unit")

    return datetime.timedelta(seconds=count * scale)


def TimespanWithLegacyFallback(text: str) -> datetime.timedelta:  # noqa: D401
    """A span of time, allowing for older style "float number of seconds" as fallback."""
    try:
        return Timespan(text)
    except ValueError:
        return datetime.timedelta(seconds=Float(text))


InfiniteTimespanType = NewType("InfiniteTimespanType", object)
InfiniteTimespan = InfiniteTimespanType(object())


def TimespanOrInfinite(text: str) -> Union[datetime.timedelta, InfiniteTimespanType]:  # noqa: D401
    """A span of time or the string 'infinite' indicating forever."""
    if text == "infinite":
        return InfiniteTimespan
    return Timespan(text)


def Percent(text: str) -> float:  # noqa: D401
    """A percentage.

    This takes a string of the form "37.2%" or "44%" and
    returns a float in the range [0.0, 1.0].

    """
    if not text.endswith("%"):
        raise ValueError("the value is not a percentage")

    percentage = float(text[:-1]) / 100.0

    if not 0 <= percentage <= 1:
        raise ValueError("percentage is out of valid range")

    return percentage


def UnixUser(text: str) -> int:  # noqa: D401
    """A Unix user name or decimal ID.

    The parsed value will be the integer user ID.

    """
    try:
        return pwd.getpwnam(text).pw_uid
    except KeyError as exc:
        try:
            return int(text)
        except ValueError:
            raise ValueError(exc) from None


def UnixGroup(text: str) -> int:  # noqa: D401
    """A Unix group name or decimal ID.

    The parsed value will be the integer group ID.

    """
    try:
        return grp.getgrnam(text).gr_gid
    except KeyError as exc:
        try:
            return int(text)
        except ValueError:
            raise ValueError(exc) from None


T = TypeVar("T")


def OneOf(**options: T) -> Callable[[str], T]:  # noqa: D401
    """One of several choices.

    For each ``option``, the name is what should be in the configuration file
    and the value is what it is mapped to.

    For example::

        OneOf(hearts="H", spades="S")

    would parse::

        "hearts"

    into::

        "H"

    """

    def one_of(text: str) -> T:
        try:
            return options[text]
        except KeyError:
            raise ValueError(f"expected one of {options.keys()!r}")

    return one_of


def TupleOf(item_parser: Callable[[str], T]) -> Callable[[str], Sequence[T]]:  # noqa: D401
    """A comma-delimited list of type T.

    At least one value must be provided. If you want an empty list
    to be a valid choice, wrap with :py:func:`Optional`.

    """

    def tuple_of(text: str) -> Sequence[T]:
        if not text:
            raise ValueError("no values provided")
        split = text.split(",")
        stripped = [item.strip() for item in split]
        return [item_parser(item) for item in stripped if item]

    return tuple_of


def DefaultFromEnv(
    item_parser: Callable[[str], T], default_src: str, fallback: OptionalType[T] = None
) -> Callable[[str], OptionalType[T]]:  # noqa: D401
    """An option of type T or a default.

    The default is sourced from an environment variable with the name specified in ``default_src``.
    If the environment variable is not set, then the fallback will be used.
    One of the following values must be provided: fallback, default_src, or the provided configuration
    """
    env = os.getenv(default_src) or ""
    default = Optional(item_parser, fallback)(env)

    def default_from_env(text: str) -> OptionalType[T]:
        val = Optional(item_parser, default)(text)
        if val:
            return val

        raise ValueError("no value provided")

    return default_from_env


def Optional(
    item_parser: Callable[[str], T], default: OptionalType[T] = None
) -> Callable[[str], OptionalType[T]]:  # noqa: D401
    """An option of type T, or ``default`` if not configured."""

    def optional(text: str) -> OptionalType[T]:
        if text:
            return item_parser(text)
        return default

    return optional


def Fallback(
    primary_parser: Callable[[str], T], fallback_parser: Callable[[str], T]
) -> Callable[[str], T]:  # noqa: D401
    """An option of type T1, or if that fails to parse, of type T2.

    This is useful for backwards-compatible configuration changes.

    """

    def fallback(text: str) -> T:
        try:
            return primary_parser(text)
        except ValueError:
            return fallback_parser(text)

    return fallback


class ConfigNamespace(dict):
    def __init__(self) -> None:
        super().__init__()
        self.__dict__ = self

    def __getattr__(self, name: str) -> Any:
        ...


ConfigSpecItem = Union["Parser", Dict[str, Any], Callable[[str], T]]
ConfigSpec = Dict[str, ConfigSpecItem]
RawConfig = Dict[str, str]


class Parser(Generic[T]):
    """Base class for configuration parsers."""

    @staticmethod
    def from_spec(spec: ConfigSpecItem) -> "Parser":
        """Return a parser for the given spec object."""
        if isinstance(spec, Parser):
            return spec
        if isinstance(spec, dict):
            return SpecParser(spec)
        if callable(spec):
            return CallableParser(spec)
        raise AssertionError(f"invalid specification: {spec!r}")

    def parse(self, key_path: str, raw_config: RawConfig) -> T:
        """Parse and return the relevant info for a given key.

        :param key_path: The key this parser is looking for.
        :param raw_config: The full raw configuration dictionary.

        """
        raise NotImplementedError


class SpecParser(Parser[ConfigNamespace]):
    """A parser that validates a static specification."""

    def __init__(self, spec: ConfigSpec):
        self.spec = spec

    def parse(self, key_path: str, raw_config: RawConfig) -> ConfigNamespace:
        parsed = ConfigNamespace()
        for key, spec in self.spec.items():
            assert "." not in key, "dots are not allowed in keys"

            if key_path:
                sub_key_path = f"{key_path}.{key}"
            else:
                sub_key_path = key

            parser = Parser.from_spec(spec)
            parsed[key] = parser.parse(sub_key_path, raw_config)
        return parsed


class CallableParser(Parser[T]):
    """A parser that wraps a simple callable."""

    def __init__(self, callable_: Callable[[str], T]):
        self.callable = callable_

    def parse(self, key_path: str, raw_config: RawConfig) -> T:
        raw_value = raw_config.get(key_path, "")

        try:
            return self.callable(raw_value)
        except Exception as exc:
            raise ConfigurationError(key_path, exc)


class DictOf(Parser[ConfigNamespace]):
    """A group of options of a given type.

    This is useful for providing data to the application without the
    application having to know ahead of time all of the possible keys.

    .. highlight:: ini

    .. include:: ../../../config_dictof_example.ini
       :literal:

    .. highlight:: py

    .. testsetup:: dictof_simple

        import configparser
        from baseplate import config
        config_parser = configparser.RawConfigParser()
        config_parser.read_file(open("docs/config_dictof_example.ini"))
        raw_config = dict(config_parser.items("app:main"))

    .. doctest:: dictof_simple

        >>> cfg = config.parse_config(raw_config, {
        ...     "population": config.DictOf(config.Integer),
        ... })

        >>> len(cfg.population)
        5

        >>> cfg.population["br"]
        207645000

    It can also be combined with other configuration specs or parsers to parse
    more complicated structures:

    .. highlight:: ini

    .. include:: ../../../config_dictof_spec_example.ini
       :literal:

    .. highlight:: py

    .. testsetup:: dictof_spec

        import configparser
        from baseplate import config
        config_parser = configparser.RawConfigParser()
        config_parser.read_file(open("docs/config_dictof_spec_example.ini"))
        raw_config = dict(config_parser.items("app:main"))

    .. doctest:: dictof_spec

        >>> cfg = config.parse_config(raw_config, {
        ...     "countries": config.DictOf({
        ...         "population": config.Integer,
        ...         "capital": config.String,
        ...     }),
        ... })

        >>> len(cfg.countries)
        5

        >>> cfg.countries["cn"].capital
        'Beijing'

        >>> cfg.countries["id"].population
        263447000

    """

    def __init__(self, spec: ConfigSpecItem):
        self.subparser = Parser.from_spec(spec)

    def parse(self, key_path: str, raw_config: RawConfig) -> ConfigNamespace:
        # match keys that start out with the prefix we expect (key_path) and
        # extract the subkey from the.key.prefix.{subkey}.the.rest
        if key_path:
            root = key_path + "."
        else:
            root = ""
        matcher = re.compile("^" + root.replace(".", r"\.") + r"([^.]+)")

        values = ConfigNamespace()
        seen_subkeys: Set[str] = set()
        for key in raw_config:
            m = matcher.search(key)
            if not m:
                continue

            subkey = m.group(1)
            if subkey in seen_subkeys:
                continue

            full_path = root + subkey
            values[subkey] = self.subparser.parse(full_path, raw_config)
            seen_subkeys.add(subkey)
        return values


def parse_config(config: RawConfig, spec: ConfigSpec) -> ConfigNamespace:
    """Parse options against a spec and return a structured representation.

    :param config: The raw stringy configuration dictionary.
    :param spec: A specification of what the configuration should look like.
    :raises: :py:exc:`ConfigurationError` The configuration violated the spec.
    :return: A structured configuration object.

    """
    parser = Parser.from_spec(spec)
    return parser.parse("", config)



from contextlib import nullcontext as does_not_raise
from unittest import mock

import pytest

from prometheus_client import REGISTRY
from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.transport.TTransport import TTransportException

from baseplate.frameworks.thrift import _ContextAwareHandler
from baseplate.frameworks.thrift import PROM_ACTIVE
from baseplate.frameworks.thrift import PROM_LATENCY
from baseplate.frameworks.thrift import PROM_REQUESTS
from baseplate.thrift.ttypes import Error
from baseplate.thrift.ttypes import ErrorCode


class Test_ThriftServerPrometheusMetrics:
    def setup(self):
        PROM_LATENCY.clear()
        PROM_REQUESTS.clear()
        PROM_ACTIVE.clear()

    @pytest.mark.parametrize(
        "exc,convert,exc_type,status,status_code,expectation",
        [
            (None, True, "", "", "", does_not_raise()),
            (
                TApplicationException(TApplicationException.UNKNOWN_METHOD, "unknown method"),
                True,
                "TApplicationException",
                "",
                "",
                pytest.raises(TApplicationException),
            ),
            (
                TProtocolException(message="Required field is unset!"),
                True,
                "TProtocolException",
                "",
                "",
                pytest.raises(TProtocolException),
            ),
            (
                TTransportException(message="Something is wrong with the transport"),
                True,
                "TTransportException",
                "",
                "",
                pytest.raises(TTransportException),
            ),
            (
                Error(ErrorCode.NOT_FOUND, "404 not found"),
                True,
                "Error",
                "NOT_FOUND",
                "404",
                pytest.raises(Error),
            ),
            (
                Error(ErrorCode.SERVICE_UNAVAILABLE, "503 unavailable"),
                True,
                "Error",
                "SERVICE_UNAVAILABLE",
                "503",
                pytest.raises(Error),
            ),
            (
                TException(message="Some other generic thrift exception"),
                True,
                "TException",
                "",
                "",
                pytest.raises(TException),
            ),
            (
                Exception("Some very generic exception"),
                False,
                "Exception",
                "",
                "",
                pytest.raises(Exception),
            ),
            (
                Exception("Some very generic exception"),
                True,
                "Error",
                "INTERNAL_SERVER_ERROR",
                "500",
                pytest.raises(Error),
            ),
        ],
    )
    def test_thrift_server_prometheus_metrics(
        self, exc, convert, exc_type, status, status_code, expectation
    ):
        class Handler:
            def handle(*args, **kwargs):
                if exc is None:
                    pass
                else:
                    raise exc

        context_handler = _ContextAwareHandler(
            handler=Handler(),
            context=mock.MagicMock(),
            logger=mock.MagicMock(),
            convert_to_baseplate_error=convert,
        )

        with expectation:
            context_handler.handle()

        common_prom_labels = {
            "thrift_method": "handle",
            "thrift_success": str(exc is None).lower(),
        }

        other_prom_labels = {
            "thrift_exception_type": exc_type,
            "thrift_baseplate_status": status,
            "thrift_baseplate_status_code": status_code,
        }

        assert (
            REGISTRY.get_sample_value(
                "thrift_server_requests_total", {**common_prom_labels, **other_prom_labels}
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                "thrift_server_latency_seconds_bucket", {**common_prom_labels, "le": "+Inf"}
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value("thrift_server_active_requests", {"thrift_method": "handle"})
            == 0
        )



import datetime
import logging
import socket

from typing import Any
from typing import Dict

from gevent.pool import Pool
from gevent.pywsgi import LoggingLogAdapter
from gevent.pywsgi import WSGIServer
from gevent.server import StreamServer

from baseplate.lib import config
from baseplate.server import _load_factory
from baseplate.server import runtime_monitor


logger = logging.getLogger(__name__)


def make_server(server_config: Dict[str, str], listener: socket.socket, app: Any) -> StreamServer:
    """Make a gevent server for WSGI apps."""
    # pylint: disable=maybe-no-member
    cfg = config.parse_config(
        server_config,
        {
            "handler": config.Optional(config.String, default=None),
            "max_concurrency": config.Optional(config.Integer),
            "stop_timeout": config.Optional(
                config.TimespanWithLegacyFallback, default=datetime.timedelta(seconds=10)
            ),
        },
    )

    if cfg.max_concurrency is not None:
        raise Exception(
            "The max_concurrency setting is not allowed for WSGI servers. See https://github.com/reddit/baseplate.py-upgrader/wiki/v1.2#max_concurrency-is-deprecated."
        )

    pool = Pool()
    log = LoggingLogAdapter(logger, level=logging.DEBUG)

    kwargs: Dict[str, Any] = {}
    if cfg.handler:
        kwargs["handler_class"] = _load_factory(cfg.handler, default_name=None)

    server = WSGIServer(
        listener,
        application=app,
        spawn=pool,
        log=log,
        error_log=LoggingLogAdapter(logger, level=logging.ERROR),
        **kwargs,
    )
    server.stop_timeout = cfg.stop_timeout.total_seconds()

    runtime_monitor.start(server_config, app, pool)
    return server



import itertools

from setuptools import find_packages
from setuptools import setup


extras_require = {
    "amqp": ["kombu>=4.0.0,<5.0"],
    "cassandra": ["cassandra-driver>=3.13.0,<4.0"],
    "cqlmapper": ["reddit-cqlmapper>=0.3.0,<1.0"],
    "kafka": ["confluent-kafka>=1.3,<2.0"],
    "memcache": ["pymemcache>=1.3.0,<1.4.4"],
    "prometheus": ["prometheus-client>=0.12.0,<1.0"],
    "pyramid": ["pyramid>=1.9.0,<2.0"],
    "redis": ["redis>=2.10.0,<4.0.0"],
    "redis-py-cluster": ["redis-py-cluster>=2.1.2,<3.0.0"],
    "refcycle": ["objgraph>=3.0,<4.0"],
    "requests": ["advocate>=1.0.0,<2.0"],
    "s3fetcher": ["boto3>=1.18"],  # Kept for backwards compatibility, these are now main requirements
    "sentry": ["sentry-sdk>=0.19,<1.0"],
    "sql": ["sqlalchemy>=1.1.0,<2.0"],
    "zookeeper": ["kazoo>=2.5.0,<3.0"],
}
extras_require["all"] = list(itertools.chain.from_iterable(extras_require.values()))

setup(
    name="baseplate",
    description="reddit's python service framework",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="reddit",
    license="BSD",
    url="https://github.com/reddit/baseplate.py",
    project_urls={
        "Documentation": "https://baseplate.readthedocs.io/en/stable/",
        "Source": "https://github.com/reddit/baseplate.py",
    },
    use_scm_version=True,
    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.7",
    setup_requires=["setuptools_scm"],
    install_requires=[
        "boto3>=1.18",
        "posix_ipc>=1.0.0,<2.0",
        "python-json-logger>=2.0,<3.0",
        "requests>=2.21.0,<3.0",
        "thrift-unofficial>=0.14.1,<1.0",
        "gevent>=20.5.0",
        "prometheus-client>=0.12.0",
    ],
    extras_require=extras_require,
    scripts=[
        "bin/baseplate-script",
        "bin/baseplate-serve",
        "bin/baseplate-shell",
        "bin/baseplate-tshell",
        "bin/baseplate-healthcheck",
    ],
    # the thrift compiler must be able to find baseplate.thrift to build
    # services which extend BaseplateService.
    package_data={"baseplate": ["py.typed"], "baseplate.thrift": ["*.thrift"]},
    zip_safe=False,
    entry_points={
        "distutils.commands": [
            "build_thrift = baseplate.frameworks.thrift.command:BuildThriftCommand"
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)
